r"""The sensorgraph subsystem is an embedded event based scripting engine for IOTile devices.

It is designed to allow you to embed a small set automatic actions that should
be run whenever events happen inside the device.  Sensorgraph is structured as
a dependency tree where actions are linked by named FIFOs that route data from
inputs -> processing functions -> output sinks.

The sensor-graph language and implementation were designed to facilitate
static analysis of maximum resource needs as well as runtime properties like
power consumption.

The sensor-graph subsystem interacts with the sensor-log and stream-manager
subsystems in the following way:

sensor-graph has rules that respond to events on a device and generate data
in named FIFOs called streams.  All of the actual data storage and stream
management on behalf of sensor-graph is handled by the sensor-log subsystem.

The streaming subsystem listens on configured streams to see if it should
package them up and send them to an external user in the form of a `report`.

So the interaction diagram looks like:


    sensor-graph       <------XXXX------>   stream-manager
                       no direct contact

    generates data in streams               sole purpose in life is to
    according to configured rules           build reports from streams

                    \\                  //
      producer       \\                //   consumer only
      and consumer    \\              //    (generates no readings!)
                          sensor-log

                          stores all streams and lets you
                          inspect their contents using
                          stream walkers that automatically
                          get updated when new data is available.

The actual work of simulating the functionality of the embedded sensor-graph
engine is performed by the iotile-sensorgraph package that makes it available
in a variety of contexts including a direct command line simulator named
iotile-sgrun as well as an optimizing compiler.

This controller subsystem mixin just wraps the underlying sensorgraph
simulator from iotile-sensorgraph and adds the correct RPC based interface to
it to emulate how you interact with sensor-graph in a physical IOTile based
device.

TODO:
  - [ ] Add dump/restore support
  - [ ] Add support for logging information on reset
"""

import logging
import struct
import asyncio
import inspect
from collections import deque
from iotile.core.hw.virtual import tile_rpc
from iotile.core.hw.exceptions import RPCErrorCode
from iotile.core.hw.reports import IOTileReading
from iotile.sg import DataStream, SensorGraph
from iotile.sg.sim.executor import RPCExecutor
from iotile.sg.node_descriptor import parse_binary_descriptor, create_binary_descriptor
from iotile.sg import streamer_descriptor
from iotile.sg.exceptions import NodeConnectionError, ProcessingFunctionError, ResourceUsageError, UnresolvedIdentifierError, StreamEmptyError
from ...constants import rpcs, pack_error, Error, ControllerSubsystem, SensorGraphError, SensorLogError
from .controller_system import ControllerSubsystemBase

def _pack_sgerror(short_code):
    """Pack a short error code with the sensorgraph subsystem."""

    return pack_error(ControllerSubsystem.SENSOR_GRAPH, short_code)


class StreamerStatus(object):
    """A model representing the state of a streamer resource."""

    def __init__(self):
        self.last_attempt_time = 0
        self.last_success_time = 0
        self.last_error = 0
        self.last_status = 0
        self.attempt_number = 0
        self.comm_status = 0


class EmulatedRPCExecutor(RPCExecutor):
    def __init__(self, device):
        super(EmulatedRPCExecutor, self).__init__()
        self.device = device
        self.logger = logging.getLogger(__name__)

    async def rpc(self, address, rpc_id):
        self.logger.debug("Sending rpc from sensorgraph to %d:%04X", address, rpc_id)

        result, = await self.device.emulator.await_rpc(address, rpc_id, bytes(), resp_format="L")
        return result


class SensorGraphSubsystem(ControllerSubsystemBase):
    """Container for sensor graph state.

    There is a distinction between which sensor-graph is saved into persisted
    storage vs currently loaded and running.  The sensor-graph subsystem runs
    a background task that receives inputs to process and processes them.
    """

    def __init__(self, sensor_log_system, stream_manager, model, emulator, executor=None):
        super(SensorGraphSubsystem, self).__init__(emulator)

        self._logger = logging.getLogger(__name__)

        self._model = model

        self._sensor_log = sensor_log_system.storage
        self._allocate_id = sensor_log_system.allocate_id
        self._inputs = emulator.create_queue(register=True)

        self._stream_manager = stream_manager
        self._rsl = sensor_log_system
        self._executor = executor

        self.graph = SensorGraph(self._sensor_log, model=model, enforce_limits=True)

        self.persisted_exists = False
        self.persisted_nodes = []
        self.persisted_streamers = []
        self.persisted_constants = []

        self.streamer_acks = {}
        self.streamer_status = {}

        self.enabled = False

        # Clock manager linkage
        self.get_timestamp = lambda: 0

    async def _reset_vector(self):
        """Background task to initialize this system in the event loop."""

        self._logger.debug("sensor_graph subsystem task starting")

        # If there is a persistent sgf loaded, send reset information.

        self.initialized.set()

        while True:
            stream, reading = await self._inputs.get()

            try:
                await process_graph_input(self.graph, stream, reading, self._executor)
                self.process_streamers()
            except:  #pylint:disable=bare-except;This is a background task that should not die
                self._logger.exception("Unhandled exception processing sensor_graph input (stream=%s), reading=%s", stream, reading)
            finally:
                self._inputs.task_done()

    def clear_to_reset(self, config_vars):
        """Clear all volatile information across a reset.

        The reset behavior is that:
        - any persisted sensor_graph is loaded
        - if there is a persisted graph found, enabled is set to True
        - if there is a persisted graph found, reset readings are pushed
          into it.
        """

        super(SensorGraphSubsystem, self).clear_to_reset(config_vars)

        self.graph.clear()

        if not self.persisted_exists:
            return

        for node in self.persisted_nodes:
            self.graph.add_node(node)

        for streamer_desc in self.persisted_streamers:
            streamer = streamer_descriptor.parse_string_descriptor(streamer_desc)
            self.graph.add_streamer(streamer)

        # Load in the constants
        for stream, reading in self.persisted_constants:
            self._sensor_log.push(stream, reading)

        self.enabled = True

        # Set up all streamers
        for index, value in self.streamer_acks.items():
            self._seek_streamer(index, value)

        #FIXME: queue sending reset readings

    def process_input(self, encoded_stream, value):
        """Process or drop a graph input.

        This method asynchronously queued an item to be processed by the
        sensorgraph worker task in _reset_vector.  It must be called from
        inside the emulation loop and returns immediately before the input is
        processed.
        """

        if not self.enabled:
            return

        if isinstance(encoded_stream, str):
            stream = DataStream.FromString(encoded_stream)
            encoded_stream = stream.encode()
        elif isinstance(encoded_stream, DataStream):
            stream = encoded_stream
            encoded_stream = stream.encode()
        else:
            stream = DataStream.FromEncoded(encoded_stream)

        reading = IOTileReading(self.get_timestamp(), encoded_stream, value)

        self._inputs.put_nowait((stream, reading))

    def _seek_streamer(self, index, value):
        """Complex logic for actually seeking a streamer to a reading_id.

        This routine hides all of the gnarly logic of the various edge cases.
        In particular, the behavior depends on whether the reading id is found,
        and if it is found, whether it belongs to the indicated streamer or not.

        If not, the behavior depends on whether the sought reading it too high
        or too low.
        """

        highest_id = self._rsl.highest_stored_id()

        streamer = self.graph.streamers[index]
        if not streamer.walker.buffered:
            return _pack_sgerror(SensorLogError.CANNOT_USE_UNBUFFERED_STREAM)

        find_type = None
        try:
            exact = streamer.walker.seek(value, target='id')
            if exact:
                find_type = 'exact'
            else:
                find_type = 'other_stream'

        except UnresolvedIdentifierError:
            if value > highest_id:
                find_type = 'too_high'
            else:
                find_type = 'too_low'

        # If we found an exact match, move one beyond it

        if find_type == 'exact':
            try:
                streamer.walker.pop()
            except StreamEmptyError:
                pass

            error = Error.NO_ERROR
        elif find_type == 'too_high':
            streamer.walker.skip_all()
            error = _pack_sgerror(SensorLogError.NO_MORE_READINGS)
        elif find_type == 'too_low':
            streamer.walker.seek(0, target='offset')
            error = _pack_sgerror(SensorLogError.NO_MORE_READINGS)
        else:
            error = _pack_sgerror(SensorLogError.ID_FOUND_FOR_ANOTHER_STREAM)

        return error

    def acknowledge_streamer(self, index, ack, force):
        """Acknowledge a streamer value as received from the remote side."""

        if index >= len(self.graph.streamers):
            return _pack_sgerror(SensorGraphError.STREAMER_NOT_ALLOCATED)

        old_ack = self.streamer_acks.get(index, 0)

        if ack != 0:
            if ack <= old_ack and not force:
                return _pack_sgerror(SensorGraphError.OLD_ACKNOWLEDGE_UPDATE)

            self.streamer_acks[index] = ack

        current_ack = self.streamer_acks.get(index, 0)
        return self._seek_streamer(index, current_ack)

    def _handle_streamer_finished(self, index, succeeded, highest_ack):
        """Callback when a streamer finishes processing."""

        self._logger.debug("Rolling back streamer %d after streaming, highest ack from streaming subsystem was %d", index, highest_ack)
        self.acknowledge_streamer(index, highest_ack, False)

    def process_streamers(self):
        """Check if any streamers should be handed to the stream manager."""

        # Check for any triggered streamers and pass them to stream manager
        in_progress = self._stream_manager.in_progress()
        triggered = self.graph.check_streamers(blacklist=in_progress)

        for streamer in triggered:
            self._stream_manager.process_streamer(streamer, callback=self._handle_streamer_finished)

    def trigger_streamer(self, index):
        """Pass a streamer to the stream manager if it has data."""

        self._logger.debug("trigger_streamer RPC called on streamer %d", index)

        if index >= len(self.graph.streamers):
            return _pack_sgerror(SensorGraphError.STREAMER_NOT_ALLOCATED)

        if index in self._stream_manager.in_progress():
            return _pack_sgerror(SensorGraphError.STREAM_ALREADY_IN_PROGRESS)

        streamer = self.graph.streamers[index]
        if not streamer.triggered(manual=True):
            return _pack_sgerror(SensorGraphError.STREAMER_HAS_NO_NEW_DATA)

        self._logger.debug("calling mark_streamer on streamer %d from trigger_streamer RPC", index)
        self.graph.mark_streamer(index)

        self.process_streamers()

        return Error.NO_ERROR

    def count_nodes(self):
        """Count the number of nodes."""

        return len(self.graph.nodes)

    def persist(self):
        """Trigger saving the current sensorgraph to persistent storage."""

        self.persisted_nodes = self.graph.dump_nodes()
        self.persisted_streamers = self.graph.dump_streamers()
        self.persisted_exists = True
        self.persisted_constants = self._sensor_log.dump_constants()

    def reset(self):
        """Clear the sensorgraph from RAM and flash."""

        self.persisted_exists = False
        self.persisted_nodes = []
        self.persisted_streamers = []
        self.persisted_constants = []
        self.graph.clear()

        self.streamer_status = {}

    def add_node(self, binary_descriptor):
        """Add a node to the sensor_graph using a binary node descriptor.

        Args:
            binary_descriptor (bytes): An encoded binary node descriptor.

        Returns:
            int: A packed error code.
        """

        try:
            node_string = parse_binary_descriptor(binary_descriptor)
        except:
            self._logger.exception("Error parsing binary node descriptor: %s", binary_descriptor)
            return _pack_sgerror(SensorGraphError.INVALID_NODE_STREAM)  # FIXME: Actually provide the correct error codes here

        try:
            self.graph.add_node(node_string)
        except NodeConnectionError:
            return _pack_sgerror(SensorGraphError.STREAM_NOT_IN_USE)
        except ProcessingFunctionError:
            return _pack_sgerror(SensorGraphError.INVALID_PROCESSING_FUNCTION)
        except ResourceUsageError:
            return _pack_sgerror(SensorGraphError.NO_NODE_SPACE_AVAILABLE)

        return Error.NO_ERROR

    def add_streamer(self, binary_descriptor):
        """Add a streamer to the sensor_graph using a binary streamer descriptor.

        Args:
            binary_descriptor (bytes): An encoded binary streamer descriptor.

        Returns:
            int: A packed error code
        """

        streamer = streamer_descriptor.parse_binary_descriptor(binary_descriptor)

        try:
            self.graph.add_streamer(streamer)
            self.streamer_status[len(self.graph.streamers) - 1] = StreamerStatus()

            return Error.NO_ERROR
        except ResourceUsageError:
            return _pack_sgerror(SensorGraphError.NO_MORE_STREAMER_RESOURCES)

    def inspect_streamer(self, index):
        """Inspect the streamer at the given index."""

        if index >= len(self.graph.streamers):
            return [_pack_sgerror(SensorGraphError.STREAMER_NOT_ALLOCATED), b'\0'*14]

        return [Error.NO_ERROR, streamer_descriptor.create_binary_descriptor(self.graph.streamers[index])]

    def inspect_node(self, index):
        """Inspect the graph node at the given index."""

        if index >= len(self.graph.nodes):
            raise RPCErrorCode(6)  #FIXME: use actual error code here for UNKNOWN_ERROR status

        return create_binary_descriptor(str(self.graph.nodes[index]))

    def query_streamer(self, index):
        """Query the status of the streamer at the given index."""

        if index >= len(self.graph.streamers):
            return None

        info = self.streamer_status[index]
        highest_ack = self.streamer_acks.get(index, 0)

        return [info.last_attempt_time, info.last_success_time, info.last_error, highest_ack, info.last_status, info.attempt_number, info.comm_status]


class SensorGraphMixin(object):
    """Mixin for an IOTileController that implements the sensor-graph subsystem.

    Args:
        sensor_log (SensorLog): The rsl subsystem.
        stream_man (StreamManager): The stream manager subsystem
        model (DeviceModel): A device model containing resource limits about the
            emulated device.
    """

    def __init__(self, emulator, sensor_log, stream_manager, model):
        self.sensor_graph = SensorGraphSubsystem(sensor_log, stream_manager, model, emulator, executor=EmulatedRPCExecutor(self._device))
        self._post_config_subsystems.append(self.sensor_graph)

    @tile_rpc(*rpcs.SG_COUNT_NODES)
    def sg_count_nodes(self):
        """Count the number of nodes in the sensor_graph."""

        return [self.sensor_graph.count_nodes()]

    @tile_rpc(*rpcs.SG_ADD_NODE)
    def sg_add_node(self, descriptor):
        """Add a node to the sensor_graph using its binary descriptor."""

        err = self.sensor_graph.add_node(descriptor)
        return [err]

    @tile_rpc(*rpcs.SG_SET_ONLINE)
    def sg_set_online(self, online):
        """Set the sensor-graph online/offline."""

        self.sensor_graph.enabled = bool(online)
        return [Error.NO_ERROR]

    @tile_rpc(*rpcs.SG_GRAPH_INPUT)
    def sg_graph_input(self, value, stream_id):
        """"Present a graph input to the sensor_graph subsystem."""

        self.sensor_graph.process_input(stream_id, value)
        return [Error.NO_ERROR]

    @tile_rpc(*rpcs.SG_RESET_GRAPH)
    def sg_reset_graph(self):
        """Clear the in-memory and persisted graph (if any)."""
        self.sensor_graph.reset()
        return [Error.NO_ERROR]

    @tile_rpc(*rpcs.SG_PERSIST_GRAPH)
    def sg_persist_graph(self):
        """Save the current in-memory graph persistently."""

        self.sensor_graph.persist()
        return [Error.NO_ERROR]

    @tile_rpc(*rpcs.SG_INSPECT_GRAPH_NODE)
    def sg_inspect_graph_node(self, index):
        """Inspect the given graph node."""

        desc = self.sensor_graph.inspect_node(index)
        return [desc]

    @tile_rpc(*rpcs.SG_ADD_STREAMER)
    def sg_add_streamer(self, desc):
        """Add a graph streamer using a binary descriptor."""

        if len(desc) == 13:
            desc += b'\0'

        err = self.sensor_graph.add_streamer(desc)
        return [err]

    @tile_rpc(*rpcs.SG_INSPECT_STREAMER)
    def sg_inspect_streamer(self, index):
        """Inspect a sensorgraph streamer by index."""

        return self.sensor_graph.inspect_streamer(index)

    @tile_rpc(*rpcs.SG_TRIGGER_STREAMER)
    def sg_trigger_streamer(self, index):
        """Manually trigger a streamer."""

        err = self.sensor_graph.trigger_streamer(index)

        return [err]

    @tile_rpc(*rpcs.SG_SEEK_STREAMER)
    def sg_seek_streamer(self, index, force, value):
        """Ackowledge a streamer."""

        force = bool(force)
        err = self.sensor_graph.acknowledge_streamer(index, value, force)
        return [err]

    @tile_rpc(*rpcs.SG_QUERY_STREAMER)
    def sg_query_streamer(self, index):
        """Query the current status of a streamer."""

        resp = self.sensor_graph.query_streamer(index)
        if resp is None:
            return [struct.pack("<L", _pack_sgerror(SensorGraphError.STREAMER_NOT_ALLOCATED))]

        return [struct.pack("<LLLLBBBx", *resp)]


async def process_graph_input(graph, stream, value, rpc_executor):
    """Process an input through this sensor graph.

    The tick information in value should be correct and is transfered
    to all results produced by nodes acting on this tick.  This coroutine
    is an asyncio compatible version of SensorGraph.process_input()

    Args:
        stream (DataStream): The stream the input is part of
        value (IOTileReading): The value to process
        rpc_executor (RPCExecutor): An object capable of executing RPCs
            in case we need to do that.
    """

    graph.sensor_log.push(stream, value)

    # FIXME: This should be specified in our device model
    if stream.important:
        associated_output = stream.associated_stream()
        graph.sensor_log.push(associated_output, value)

    to_check = deque([x for x in graph.roots])

    while len(to_check) > 0:
        node = to_check.popleft()
        if node.triggered():
            try:
                results = node.process(rpc_executor, graph.mark_streamer)
                for result in results:
                    if inspect.iscoroutine(result.value):
                        result.value = await asyncio.ensure_future(result.value)

                    result.raw_time = value.raw_time
                    graph.sensor_log.push(node.stream, result)
            except:
                logging.getLogger(__name__).exception("Unhandled exception in graph node processing function for node %s", str(node))

            # If we generated any outputs, notify our downstream nodes
            # so that they are also checked to see if they should run.
            if len(results) > 0:
                to_check.extend(node.outputs)



"""Mixin for device updating via signed scripts."""

import base64
from iotile.core.hw.virtual import tile_rpc
from ...virtual import SerializableState
from ...constants import RunLevel, TileState, rpcs
from .controller_system import ControllerSubsystemBase


class TileInfo(SerializableState):
    def __init__(self, hw_type, name, api_info, fw_info, exec_info, slot, unique_id, state=TileState.INVALID, address=None):
        super(TileInfo, self).__init__()

        self.hw_type = hw_type
        self.name = name
        self.api_info = api_info
        self.fw_info = fw_info
        self.exec_info = exec_info
        self.slot = slot
        self.unique_id = unique_id

        self.state = state
        self.address = address

        self.mark_complex('name', self._dump_name, self._restore_name)

    def _dump_name(self, name):
        return base64.b64encode(name).decode('utf-8')

    def _restore_name(self, name):
        return base64.b64decode(name)

    def registration_packet(self):
        """Serialize this into a tuple suitable for returning from an RPC.

        Returns:
            tuple: The serialized values.
        """

        return (self.hw_type, self.api_info[0], self.api_info[1], self.name, self.fw_info[0], self.fw_info[1], self.fw_info[2],
                self.exec_info[0], self.exec_info[0], self.exec_info[0], self.slot, self.unique_id)

    @classmethod
    def CreateInvalid(cls):
        """Create a blank invalid TileInfo structure.

        Returns:
            TileInfo
        """

        return TileInfo(0, '\0'*6, (0, 0), (0, 0, 0), (0, 0, 0), 0, 0)


class TileManagerState(SerializableState, ControllerSubsystemBase):
    """Serializeable state object for all tile_manager state."""

    def __init__(self, emulator):
        SerializableState.__init__(self)
        ControllerSubsystemBase.__init__(self, emulator)

        self.registered_tiles = []
        self.safe_mode = False
        self.debug_mode = False
        self.queue = emulator.create_queue(register=True)

        self.mark_ignored('initialized')
        self.mark_ignored('queue')
        self.mark_complex('registered_tiles', self._dump_registered_tiles, self._restore_registered_tiles)

    async def _reset_vector(self):
        self.initialized.set()

        while True:
            info, config_rpcs = await self.queue.get()

            try:
                for rpc in config_rpcs:
                    await self._emulator.await_rpc(*rpc)

                await self._emulator.await_rpc(info.address, rpcs.START_APPLICATION)

                info.state = TileState.RUNNING
            finally:
                self.queue.task_done()

    def clear_to_reset(self, config_vars):
        """Clear to the state immediately after a reset."""

        super(TileManagerState, self).clear_to_reset(config_vars)
        self.registered_tiles = self.registered_tiles[:1]
        self.safe_mode = False
        self.debug_mode = False

    def _dump_registered_tiles(self, tiles):
        return [info.dump() for info in tiles]

    def _restore_registered_tiles(self, serialized_tiles):
        tiles = []

        for serialized_tile in serialized_tiles:
            tile = TileInfo(None, None, None, None, None, None, None)
            tile.restore(serialized_tile)

            tiles.append(tile)

        return tiles

    def insert_tile(self, tile_info):
        """Add or replace an entry in the tile cache.

        Args:
            tile_info (TileInfo): The newly registered tile.
        """

        for i, tile in enumerate(self.registered_tiles):
            if tile.slot == tile_info.slot:
                self.registered_tiles[i] = tile_info
                return

        self.registered_tiles.append(tile_info)


class TileManagerMixin(object):
    """Reference controller subsystem for device updating.

    This class must be used as a mixin with a ReferenceController base class.
    """


    def __init__(self, emulator):
        self.tile_manager = TileManagerState(emulator)

        # Register the controller itself into our tile_info database
        info = TileInfo(self.hardware_type, self.name, self.api_version, self.firmware_version, self.executive_version, 0, 0, state=TileState.RUNNING)
        self.tile_manager.insert_tile(info)

        self._post_config_subsystems.append(self.tile_manager)

    @tile_rpc(*rpcs.REGISTER_TILE)
    def register_tile(self, hw_type, api_major, api_minor, name, fw_major, fw_minor, fw_patch, exec_major, exec_minor, exec_patch, slot, unique_id):
        """Register a tile with this controller.

        This function adds the tile immediately to its internal cache of registered tiles
        and queues RPCs to send all config variables and start tile rpcs back to the tile.
        """

        api_info = (api_major, api_minor)
        fw_info = (fw_major, fw_minor, fw_patch)
        exec_info = (exec_major, exec_minor, exec_patch)

        address = 10 + slot
        info = TileInfo(hw_type, name, api_info, fw_info, exec_info, slot, unique_id, state=TileState.JUST_REGISTERED, address=address)

        self.tile_manager.insert_tile(info)

        debug = int(self.tile_manager.debug_mode)

        if self.tile_manager.safe_mode:
            run_level = RunLevel.SAFE_MODE
            info.state = TileState.SAFE_MODE
            config_rpcs = []
        else:
            run_level = RunLevel.START_ON_COMMAND
            info.state = TileState.BEING_CONFIGURED
            config_rpcs = self.config_database.stream_matching(address, name)

        self.tile_manager.queue.put_nowait((info, config_rpcs))

        return [address, run_level, debug]

    @tile_rpc(*rpcs.COUNT_TILES)
    def count_tiles(self):
        """Count the number of registered tiles including the controller."""

        return [len(self.tile_manager.registered_tiles)]

    @tile_rpc(*rpcs.DESCRIBE_TILE)
    def describe_tile(self, index):
        """Get the registration information for the tile at the given index."""

        if index >= len(self.tile_manager.registered_tiles):
            tile = TileInfo.CreateInvalid()
        else:
            tile = self.tile_manager.registered_tiles[index]

        return tile.registration_packet()



"""Custom exceptions for iotile-emulate."""

import struct
from iotile.core.exceptions import InternalError
from ..constants import pack_error

class RPCRuntimeError(Exception):
    """Exception that indicates an RPC had a runtime error.

    Runtime errors are application defined and are translated
    to a 16-bit or 32-bit return value containing the error
    code.

    Args:
        code (int): The error code that should be returned from
            the RPC.
        subsystem (int): The subsystem that the error code should
            be returned from.  This may only be passed if size=L
        size (str): Whether to return a 32-bit long-form error
            including subsystem or a 16-bit short-firm error.
            You may pass L or H.  The default is L, meaning
            32 bit errors.
    """

    def __init__(self, code, subsystem=None, size="L"):
        super(RPCRuntimeError, self).__init__()

        if size not in ("L", "H"):
            raise InternalError("Unknown size {} in RPCRuntimeError".format(size))

        if subsystem is not None and size == 'H':
            raise InternalError("You cannot combine size=H with a subsystem, subsystem={}".format(subsystem))

        if subsystem is not None:
            error = pack_error(subsystem, code)
        else:
            error = code

        self.code = error
        self.subsystem = subsystem
        self.packed_error = error

        print("Packed error: %r" % self.packed_error)
        self.binary_error = struct.pack("<{}".format(size), self.packed_error)



"""Test coverage of the ReferenceDevice and ReferenceController emulated objects."""

import sys
import pytest
from iotile.core.hw import HardwareManager
from iotile.core.exceptions import HardwareError
from iotile.emulate.virtual import EmulatedPeripheralTile
from iotile.emulate.reference import ReferenceDevice
from iotile.emulate.constants import rpcs, Error
from iotile.emulate.transport import EmulatedDeviceAdapter


@pytest.fixture(scope="function")
def reference():
    """Get a reference device with a controller and single peripheral tile."""

    device = ReferenceDevice({'simulate_time': False})
    peripheral = EmulatedPeripheralTile(10, device)
    device.add_tile(10, peripheral)

    device.start()
    yield device, peripheral
    device.stop()


@pytest.fixture(scope="function")
def reference_hw():
    """Get a reference device and connected HardwareManager."""

    device = ReferenceDevice({'simulate_time': False})
    peripheral = EmulatedPeripheralTile(11, device)
    peripheral.declare_config_variable("test 1", 0x8000, 'uint16_t')
    peripheral.declare_config_variable('test 2', 0x8001, 'uint32_t[5]')

    device.add_tile(11, peripheral)

    adapter = EmulatedDeviceAdapter(None, devices=[device])

    with HardwareManager(adapter=adapter) as hw:
        hw.connect(1)

        yield hw, device, peripheral


def test_basic_usage():
    """Make sure we can import and use the objects."""

    with HardwareManager(port='emulated:reference_1_0@#eyJzaW11bGF0ZV90aW1lIjogZmFsc2V9') as hw:
        hw.connect(1)
        debug = hw.debug()

        state = debug.dump_snapshot()
        debug.restore_snapshot(state)


def test_peripheral_tiles():
    """Make sure the controller tile properly brings up the peripheral tiles in a controlled manner."""

    # Don't use the fixture since the purpose of this test is to make sure the fixture works
    device = ReferenceDevice({'simulate_time': False})
    peripheral = EmulatedPeripheralTile(10, device)
    device.add_tile(10, peripheral)

    device.start()
    device.stop()


def test_config_variable_rpcs(reference):
    """Make sure that we can set, list and get config variables."""

    device, peripheral = reference

    peripheral.declare_config_variable("test 1", 0x8000, 'uint16_t')
    peripheral.declare_config_variable('test 2', 0x8001, 'uint32_t[5]')

    # Test listing
    resp = device.simple_rpc(10, rpcs.LIST_CONFIG_VARIABLES, 0)
    count = resp[0]
    ids = resp[1:]

    assert len(ids) == 9
    assert count == 2
    assert ids[:2] == (0x8000, 0x8001)

    # Test describing
    resp1 = device.simple_rpc(10, rpcs.DESCRIBE_CONFIG_VARIABLE, 0x8000)
    resp2 = device.simple_rpc(10, rpcs.DESCRIBE_CONFIG_VARIABLE, 0x8001)
    resp3 = device.simple_rpc(10, rpcs.DESCRIBE_CONFIG_VARIABLE, 0x8002)

    assert len(resp1) == 5
    assert len(resp2) == 5
    assert len(resp3) == 5
    assert resp3[0] == Error.INVALID_ARRAY_KEY

    err, _, _, config_id, packed_size = resp1
    assert err == 0
    assert config_id == 0x8000
    assert packed_size == 2

    err, _, _, config_id, packed_size = resp2
    assert err == 0
    assert config_id == 0x8001
    assert packed_size == (1 << 15) | (5*4)

    # Test setting (make sure to artificially force pre-app started state)
    peripheral.initialized.clear()
    err, = device.simple_rpc(10, rpcs.SET_CONFIG_VARIABLE, 0x8000, 0, bytes(bytearray([5, 6])))
    assert err == 0

    err, = device.simple_rpc(10, rpcs.SET_CONFIG_VARIABLE, 0x8001, 0, bytes(bytearray([7, 8])))
    assert err == 0

    err, = device.simple_rpc(10, rpcs.SET_CONFIG_VARIABLE, 0x8001, 19, bytes(bytearray([7, 8])))
    assert err == Error.INPUT_BUFFER_TOO_LONG

    err, = device.simple_rpc(10, rpcs.SET_CONFIG_VARIABLE, 0x8002, 0, bytes(bytearray([7, 8])))
    assert err == Error.INVALID_ARRAY_KEY

    peripheral.initialized.set()
    err, = device.simple_rpc(10, rpcs.SET_CONFIG_VARIABLE, 0x8000, 0, bytes(bytearray([5, 6])))
    assert err == Error.STATE_CHANGE_AT_INVALID_TIME

    # Test Getting
    data, = device.simple_rpc(10, rpcs.GET_CONFIG_VARIABLE, 0x8000, 0)
    assert data == bytearray([5, 6])

    data, = device.simple_rpc(10, rpcs.GET_CONFIG_VARIABLE, 0x8001, 0)
    assert data == bytearray([7, 8])

    data, = device.simple_rpc(10, rpcs.GET_CONFIG_VARIABLE, 0x8002, 0)
    assert data == bytearray([])


def test_snapshot_save_load(reference):
    """Make sure snapshot saving and loading works without error on a peripheral tile."""

    device, _peripheral = reference

    state = device.dump_state()
    device.restore_state(state)


def test_config_database(reference_hw):
    """Test the controller config database RPCs using the canonical proxy plugin."""

    hw, _device, _peripheral = reference_hw

    con = hw.get(8, basic=False)
    slot = hw.get(11, basic=True)
    config_db = con.config_database()

    # Make sure all of the basic RPCs work
    assert config_db.count_variables() == 0
    config_db.clear_variables()

    # Make sure we can set and get config variables
    config_db.set_variable('slot 1', 0x8000, 'uint16_t', 15)
    config_db.set_variable('slot 1', 0x8001, 'uint32_t[]', "[5, 6, 7]")

    assert config_db.count_variables() == 2
    var8000 = config_db.get_variable(1)
    var8001 = config_db.get_variable(2)

    var8000['metadata'] = str(var8000['metadata'])
    var8001['metadata'] = str(var8001['metadata'])

    assert var8000 == {
        'metadata': 'Target: slot 1\nData Offset: 0\nData Length: 4\nValid: True',
        'name': 0x8000,
        'data': bytearray([15, 0])
    }
    assert var8001 == {
        'metadata': 'Target: slot 1\nData Offset: 4\nData Length: 14\nValid: True',
        'name': 0x8001,
        'data': bytearray([5, 0, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0])
    }

    slot_conf = slot.config_manager()

    # Make sure we stream all variables on a reset
    slot.reset(wait=0)
    assert slot_conf.get_variable(0x8000) == bytearray([15, 0])
    assert slot_conf.get_variable(0x8001) == bytearray([5, 0, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0])

    # Make sure invalidated variables are not streamed and the tile resets them on reset
    config_db.invalidate_variable(1)
    assert config_db.count_variables() == 2
    slot.reset(wait=0)
    assert slot_conf.get_variable(0x8000) == bytearray()
    assert slot_conf.get_variable(0x8001) == bytearray([5, 0, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0])

    # Check memory limits before and after compaction
    usage = config_db.memory_limits()
    assert usage == {
        'data_usage': 18,
        'data_compactable': 4,
        'data_limit': 4096,
        'entry_usage': 2,
        'entry_compactable': 1,
        'entry_limit': 255
    }

    # Make sure we remove invalid and keep valid entries upon compaction
    config_db.compact_database()
    usage = config_db.memory_limits()
    assert usage == {
        'data_usage': 14,
        'data_compactable': 0,
        'data_limit': 4096,
        'entry_usage': 1,
        'entry_compactable': 0,
        'entry_limit': 255
    }

    assert config_db.count_variables() == 1
    slot.reset(wait=0)
    assert slot_conf.get_variable(0x8000) == bytearray()
    assert slot_conf.get_variable(0x8001) == bytearray([5, 0, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0])


def test_tile_manager(reference_hw):
    """Test the controller tile_manager RPCs using the canonical proxy plugin."""

    # Work around inconsistency in output of iotile-support-lib-controller-3 on python 2 and 3
    if sys.version_info.major < 3:
        con_str = 'refcn1, version 1.0.0 at slot 0'
        peri_str = 'noname, version 1.0.0 at slot 1'
    else:
        con_str = "b'refcn1', version 1.0.0 at slot 0"
        peri_str = "b'noname', version 1.0.0 at slot 1"

    hw, _device, _peripheral = reference_hw

    con = hw.get(8, basic=False)
    slot = hw.get(11, basic=True)
    tileman = con.tile_manager()

    assert tileman.count_tiles() == 2

    con = tileman.describe_tile(0)
    peri = tileman.describe_tile(1)
    assert str(con) == con_str
    assert str(peri) == peri_str

    # Make sure reseting a tile doesn't add a new entry
    slot.reset(wait=0)

    assert tileman.count_tiles() == 2

    con = tileman.describe_tile(0)
    peri = tileman.describe_tile(1)
    assert str(con) == con_str
    assert str(peri) == peri_str

    con = tileman.describe_selector('controller')
    peri = tileman.describe_selector('slot 1')
    assert str(con) == con_str
    assert str(peri) == peri_str


def test_raw_sensor_log(reference_hw):
    """Test to ensure that the raw sensor log works."""

    hw, _device, _peripheral = reference_hw

    con = hw.get(8, basic=False)
    sensor_graph = con.sensor_graph()

    assert sensor_graph.count_readings() == {
        'streaming': 0,
        'storage': 0
    }

    assert sensor_graph.highest_id() == 0

    sensor_graph.push_many('output 1', 10, 20)
    sensor_graph.push_many('buffered 1', 15, 25)

    assert sensor_graph.highest_id() == 45

    readings = sensor_graph.download_stream('output 1')
    assert len(readings) == 20

    readings = sensor_graph.download_stream('buffered 1')
    assert len(readings) == 25

    readings = sensor_graph.download_stream('buffered 1', reading_id=44)
    assert len(readings) == 2

    sensor_graph.clear()
    assert sensor_graph.count_readings() == {
        'streaming': 1,
        'storage': 0
    }

    assert sensor_graph.highest_id() == 46


def test_rsl_reset_config(reference_hw):
    """Make sure we properly load config variables into the sensor_log."""

    hw, _device, _peripheral = reference_hw

    import logging
    logger = logging.getLogger(__name__)

    con = hw.get(8, basic=False)
    sensor_graph = con.sensor_graph()
    config = con.config_database()

    config.set_variable('controller', 0x2004, 'uint8_t', 1)

    logger.info("Sending Reset 1")
    con.reset(wait=0)
    logger.info("Done with Reset 1")

    with pytest.raises(HardwareError):
        sensor_graph.push_many('buffered 1', 15, 20000)

    assert sensor_graph.count_readings() == {
        'streaming': 0,
        'storage': 16128
    }

    config.set_variable('controller', 0x2005, 'uint8_t', 1)
    logger.info("Sending Reset 2")
    con.reset(wait=0)

    with pytest.raises(HardwareError):
        sensor_graph.push_many('output 1', 15, 50000)

    assert sensor_graph.count_readings() == {
        'streaming': 48896,
        'storage': 16128
    }


def test_rsl_dump_restore(reference_hw):
    """Make sure the rsl state is properly saved and restored."""

    hw, device, _peripheral = reference_hw
    refcon = device.controller

    con = hw.get(8, basic=False)
    sensor_graph = con.sensor_graph()

    # Verify that we properly restore data and our download walker
    for i in range(0, 10):
        sensor_graph.push_reading('output 1', i)

    con.rpc(0x20, 0x08, 0x5001, result_format="LLLL")
    err, _timestamp, reading, unique_id, _act_stream = con.rpc(0x20, 0x09, 1, arg_format='B', result_format="LLLLH2x")
    assert err == 0
    assert reading == 0
    assert unique_id == 1

    state = device.dump_state()
    device.restore_state(state)

    # Make sure we keep our same place in the dump stream line
    err, _timestamp, reading, unique_id, _act_stream = con.rpc(0x20, 0x09, 1, arg_format='B', result_format="LLLLH2x")

    assert err == 0
    assert reading == 1
    assert unique_id == 2

    # Make sure we have all of our data
    assert refcon.sensor_log.engine.count() == (0, 10)

    readings = sensor_graph.download_stream('output 1')
    assert len(readings) == 10
    for i, reading in enumerate(readings):
        assert reading.value == i
        assert reading.reading_id == i + 1



"""Demo EmulatedDevice and EmulatedTile objects for testing and demonstration purposes."""

from .demo_device import DemoEmulatedDevice
from .demo_reference import DemoReferenceDevice
from .demo_proxy import DemoTileProxy

__all__ = ['DemoEmulatedDevice', 'DemoReferenceDevice', 'DemoTileProxy']



"""Base class for virtual devices designed to emulate physical devices."""

import logging
from iotile.core.exceptions import DataError
from iotile.core.utilities import SharedLoop
from iotile.core.hw.virtual import StandardVirtualDevice, pack_rpc_payload, unpack_rpc_payload
from iotile.core.hw.exceptions import AsynchronousRPCResponse, BusyRPCResponse
from .emulation_mixin import EmulationMixin
from .state_log import EmulationStateLog
from ..constants.rpcs import RPCDeclaration
from ..internal import EmulationLoop
from ..utilities import format_rpc


#pylint:disable=abstract-method;This is an abstract base class
class EmulatedDevice(EmulationMixin, StandardVirtualDevice):
    """Base class for virtual devices designed to emulate physical devices.

    This class adds additional state and test scenario loading functionality
    as well as tracing of state changes on the emulated device for comparison
    and verification purposes.

    Since all IOTile devices have a single TileBus interface that serializes
    RPCs between tiles so that only one runs at a time, this class spawns a
    single background rpc to work through the queue of RPCs that should be
    sent, sending them one at a time.  The rpc dispatch queue is started when
    start() is called and stopped synchronously when stop() is called.

    Args:
        iotile_id (int): A 32-bit integer that specifies the globally unique ID
            for this IOTile device.
        loop (BackgroundEventLoop): The loop to use to run this device's simulation.
    """

    __NO_EXTENSION__ = True

    def __init__(self, iotile_id, *, loop=SharedLoop):
        self.state_history = EmulationStateLog()

        StandardVirtualDevice.__init__(self, iotile_id)
        EmulationMixin.__init__(self, None, self.state_history)

        self._logger = logging.getLogger(__name__)
        self.emulator = EmulationLoop(self._dispatch_rpc, loop=loop)

    async def _dispatch_rpc(self, address, rpc_id, arg_payload):
        """Background work queue handler to dispatch RPCs."""

        if self.emulator.is_tile_busy(address):
            self._track_change('device.rpc_busy_response', (address, rpc_id, arg_payload, None, None), formatter=format_rpc)
            raise BusyRPCResponse()

        try:
            # Send the RPC immediately and wait for the response
            resp = await super(EmulatedDevice, self).async_rpc(address, rpc_id, arg_payload)
            self._track_change('device.rpc_sent', (address, rpc_id, arg_payload, resp, None), formatter=format_rpc)

            return resp
        except AsynchronousRPCResponse:
            self._track_change('device.rpc_started', (address, rpc_id, arg_payload, None, None), formatter=format_rpc)
            raise
        except Exception as exc:
            self._track_change('device.rpc_exception', (address, rpc_id, arg_payload, None, exc), formatter=format_rpc)
            raise

    def finish_async_rpc(self, address, rpc_id, response):
        """Finish a previous asynchronous RPC.

        This method should be called by a peripheral tile that previously
        had an RPC called on it and chose to response asynchronously by
        raising ``AsynchronousRPCResponse`` in the RPC handler itself.

        The response passed to this function will be returned to the caller
        as if the RPC had returned it immediately.

        The rpc response will be sent in the RPC thread.  By default this
        method will block until the response is finished.  If you don't
        want to block, you can pass sync=False

        Args:
            address (int): The tile address the RPC was called on.
            rpc_id (int): The ID of the RPC that was called.
            response (bytes): The bytes that should be returned to
                the caller of the RPC.
        """

        try:
            self.emulator.finish_async_rpc(address, rpc_id, response)
            self._track_change('device.rpc_finished', (address, rpc_id, None, response, None), formatter=format_rpc)
        except Exception as exc:
            self._track_change('device.rpc_exception', (address, rpc_id, None, response, exc), formatter=format_rpc)
            raise

    def start(self, channel=None):
        """Start this emulated device.

        This triggers the controller to call start on all peripheral tiles in
        the device to make sure they start after the controller does and then
        it waits on each one to make sure they have finished initializing
        before returning.

        Args:
            channel (AbstractAsyncDeviceChannel): A channel to allow pushing
                events to the client asynchronously.
        """

        super(EmulatedDevice, self).start(channel)
        self.emulator.start()

    def stop(self):
        """Stop running this virtual device including any worker threads."""

        self.emulator.stop()
        super(EmulatedDevice, self).stop()

    def dump_state(self):
        """Dump the current state of this emulated object as a dictionary.

        Returns:
            dict: The current state of the object that could be passed to load_state.
        """

        state = {}

        state['tile_states'] = {}

        for address, tile in self._tiles.items():
            state['tile_states'][address] = tile.dump_state()

        return state

    async def async_rpc(self, address, rpc_id, payload=b""):
        """Call an RPC by its address and ID.

        This will send the RPC to the background rpc dispatch thread and
        synchronously wait for the response.

        Args:
            address (int): The address of the mock tile this RPC is for
            rpc_id (int): The number of the RPC
            payload (bytes): A byte string of payload parameters up to 20 bytes

        Returns:
            bytes: the response payload from the RPC
        """

        return await self.emulator.call_rpc_internal(address, rpc_id, payload)

    def synchronize_task(self, func, *args, **kwargs):
        """Run callable in the rpc thread and wait for it to finish.

        The callable ``func`` will be passed into the EmulationLoop and run
        there.  This method will block until ``func`` is finished and
        return/raise whatever that callable returns/raises.

        This method is mainly useful for performing an activity that needs to
        be synchronized with the rpc thread for safety reasons.

        If this method is called from the rpc thread itself, it will just
        run the task and return its result.

        Args:
            func (callable): A method with signature callable(*args, **kwargs),
                that will be called with the optional *args and **kwargs passed
                to this method.
            *args: Arguments that will be passed to callable.
            **kwargs: Keyword arguments that will be passed to callable.

        Returns:
            object: Whatever callable returns after it runs.
        """

        if self.emulator.on_emulation_thread():
            return func(*args, **kwargs)

        async def _runner():
            return func(*args, **kwargs)

        return self.emulator.run_task_external(_runner())

    def wait_idle(self):
        """Wait until the emulated device is idle.

        An idle device is one where there are no tasks pending on the
        emulation thread. This is different from wait_deferred_rpcs() a given
        rpc that was scheduled before your call to wait_deferred_rpcs() could
        have scheduled an rpc to run after your call.

        This method gives you a deterministic way to wait until there is no
        more activity on the device.  This is particularly useful for
        deterministic testing because you can start an action and know
        precisely when all potential chain reactions have stopped.

        The behavior of this function depends on whether it is called within
        or outside of the emulation loop.  If it is called within the emulation
        loop then it returns an awaitable object so the correct usage is:
        ``await wait_idle()``.

        If it is called outside of the event loop it blocks the calling thread
        synchronously until the device is idle and returns None.

        This method is safe to call either inside or outside of the emulation
        loop.  However, if you are calling it inside of the emulation loop,
        you need to make sure you are not creating a deadlock by waiting for
        an idle condition inside of a handler function that must finish before
        the device can be idle.  It is not possible to deadlock with this
        function if called from outside of the event loop.
        """

        return self.emulator.wait_idle()

    def restore_state(self, state):
        """Restore the current state of this emulated device.

        Args:
            state (dict): A previously dumped state produced by dump_state.
        """

        tile_states = state.get('tile_states', {})

        for address, tile_state in tile_states.items():
            address = int(address)
            tile = self._tiles.get(address)
            if tile is None:
                raise DataError("Invalid dumped state, tile does not exist at address %d" % address, address=address)

            tile.restore_state(tile_state)

    def load_metascenario(self, scenario_list):
        """Load one or more scenarios from a list.

        Each entry in scenario_list should be a dict containing at least a
        name key and an optional tile key and args key.  If tile is present
        and its value is not None, the scenario specified will be loaded into
        the given tile only.  Otherwise it will be loaded into the entire
        device.

        If the args key is specified is will be passed as keyword arguments
        to load_scenario.

        Args:
            scenario_list (list): A list of dicts for each scenario that should
                be loaded.
        """

        for scenario in scenario_list:
            name = scenario.get('name')
            if name is None:
                raise DataError("Scenario in scenario list is missing a name parameter", scenario=scenario)

            tile_address = scenario.get('tile')
            args = scenario.get('args', {})

            dest = self
            if tile_address is not None:
                dest = self._tiles.get(tile_address)

                if dest is None:
                    raise DataError("Attempted to load a scenario into a tile address that does not exist", address=tile_address, valid_addresses=list(self._tiles))

            dest.load_scenario(name, **args)



"""A list of changes to an emulated device for verification purposes."""

import sys
import threading
from collections import namedtuple
import csv
from time import monotonic

StateChange = namedtuple("StateChange", ['time', 'tile', 'property', 'value', 'string_value'])


class EmulationStateLog(object):
    """A thread safe list of state changes to an emulated device."""

    def __init__(self):
        self.changes = []
        self.tracking = False
        self._lock = threading.Lock()
        self._whitelist = set()

    def track_change(self, tile, property_name, value, formatter=None):
        """Record that a change happened on a given tile's property.

        This will as a StateChange object to our list of changes if we
        are recording changes, otherwise, it will drop the change.

        Args:
            tile (int): The address of the tile that the change happened on.
            property_name (str): The name of the property that changed.
            value (object): The new value assigned to the property.
            formatter (callable): Optional function to convert value to a
                string.  This function will only be called if track_changes()
                is enabled and `name` is on the whitelist for properties that
                should be tracked.  If `formatter` is not passed or is None,
                it will default to `str`.
        """

        if not self.tracking:
            return

        if len(self._whitelist) > 0 and (tile, property_name) not in self._whitelist:
            return

        if formatter is None:
            formatter = str

        change = StateChange(monotonic(), tile, property_name, value, formatter(value))

        with self._lock:
            self.changes.append(change)

    def enable(self):
        """Start tracking changes."""

        self.tracking = True

    def set_whitelist(self, whitelist):
        """Only record changes to the given list of (tile, name) properties.

        Adding a whitelist allows you to ignore changes that you don't care
        about.  The argument should be a list of tuples with 2 members:
        (tile_adress, property_name)

        Once specified, only property changes in the whitelist will be recorded.

        You can clear a whitelist by calling `set_whitelist([])`

        Args:
            whitelist (list of (tile, property_name)): A list of all properties that
                should be included in the emulation state log.
        """

        self._whitelist = set(whitelist)

    def disable(self):
        """Stop tracking changes."""

        self.tracking = False

    def dump(self, out_path, header=True):
        """Save this list of changes as a csv file at out_path.

        The format of the output file will be a CSV with 4 columns:
        timestamp, tile address, property, string_value

        There will be a single header row starting the CSV output unless
        header=False is passed.

        Args:
            out_path (str): The path where we should save our current list of
                changes.
            header (bool): Whether we should include a header row in the csv
                file.  Defaults to True.
        """

        # See https://stackoverflow.com/a/3348664/9739119 for why this is necessary
        if sys.version_info[0] < 3:
            mode = "wb"
        else:
            mode = "w"

        with open(out_path, mode) as outfile:
            writer = csv.writer(outfile, quoting=csv.QUOTE_MINIMAL)
            if header:
                writer.writerow(["Timestamp", "Tile Address", "Property Name", "Value"])

            for entry in self.changes:
                writer.writerow([entry.time, entry.tile, entry.property, entry.string_value])



"""
Tests that an iotile rpcs return BusyResponse when the iotile is already
handling asynch rpc

Test includes a device with two tiles (see device.py):
* Tile 1 has an RPC Y that when called triggers it to send an RPC X to tile 2
    from a background task
* Tile 2 implements RPC X asynchronously and does not complete it until it
    gets a specific flag set
* Tile 1 and 2 also implement two other RPCs that are synchronous and
    let the proxy poll them to determine if they will return busy.
"""

import pytest
import os
from iotile.core.hw.exceptions import BusyRPCResponse
from iotile.core.hw.hwmanager import HardwareManager
from iotile.core.dev.registry import ComponentRegistry

path = os.path.join(os.path.dirname(__file__), 'busyresponse/device.py')

@pytest.fixture
def register_proxies():
    reg = ComponentRegistry()
    reg.clear()
    reg.clear_components()
    reg.register_extension('iotile.proxy', 'test_proxies', path)


@pytest.fixture
def hw():
    hw = HardwareManager(port='virtual:{}'.format(path))
    hw.connect_direct(1)

    yield hw

    hw.disconnect()
    hw.close()


def test_sync_request_during_async(register_proxies, hw):
    """
    1. Proxy calls RPC Y to start a long running background RPC
    2. Proxy then calls another synchronous RPC on tile 1 which succeeds
    3. Proxy then calls another synchronous RPC on tile 2 which should fail with busy
    4. Test case then sets flag on tile 2 to tell it to finish the async rpc
    5. Proxy then calls another synchronous RPC on tile 2 which should succeed
    """

    proxy1 = hw.get(11)
    assert proxy1.sync_rpc() == b'sync rpc tile01'

    proxy2 = hw.get(12)
    assert proxy2.sync_rpc() == b'sync rpc tile02'

    # Step 1
    proxy1.rpc_y()

    # Step 2
    assert proxy1.sync_rpc() == b'sync rpc tile01'

    # Step 3
    with pytest.raises(BusyRPCResponse):
        proxy2.sync_rpc()

    # Step 4
    proxy1.set_event()

    # Step 5
    assert proxy2.sync_rpc() == b'sync rpc tile02'



"""Main class where all emulation takes place."""

import sys
import logging
import asyncio

from iotile.core.utilities import SharedLoop
from iotile.core.exceptions import ArgumentError, InternalError, TimeoutExpiredError
from iotile.core.hw.virtual import unpack_rpc_payload, pack_rpc_payload

from .response import CrossThreadResponse, AwaitableResponse
from .rpc_queue import RPCQueue
from ..constants.rpcs import RPCDeclaration
from ..common import RPCRuntimeError

if sys.version_info < (3, 5):
    raise ImportError("EmulationLoop is only supported on python 3.5 and above")


class EmulationLoop:
    """A background thread that runs an event loop emulating a device.

    The loop is started when start() is called and cleanly stopped when stop()
    is called.  Coroutines can be added to the event loop that are tracked and
    can be stopped to simulate a device reset.

    The loop is executed on a background thread so that you can control the
    loop from the main thread.  Since this loop is designed for simulating an
    IOTile device as part of a larger emulator, it naturally has support for
    invoking RPCs inside the loop.  You must provide a handler function that
    actually dispatches the rpc.  Your handler will always be run inside the
    event loop.  There are two ways to invoke the rpcs, `await_rpc` and
    `call_rpc_external`.

    `await_rpc` must always be called from a coroutine inside the event loop
    and returns an awaitable. `call_rpc_external` must never be called from
    the event loop and is designed to be used by external callers to inject
    RPCs into the emulation.

    Unlike a normal asyncio.EventLoop, which does not have an externally
    visible concept of being idle, there is a method `wait_idle()` on
    EmulationLoop that will block until the emulator is idle.  Idleness is
    defined as when all internal workqueues inside the emulated device are
    empty.  This includes having no pending RPCs as a base condition, but any
    tile can also register work queues that contain background work and must
    be empty for the emulator to be considered idle.

    The ``wait_idle()`` method is the primary way that external users are able
    to synchronously interact with the EmulationLoop.  They can apply a
    stimulus like sending and RPC and then call wait_idle() to wait until all
    of the ripples of the RPC have settled down.  This allows for writing
    simple synchronous code that interacts with the EmulationLoop externally.

    Args:
        rpc_handler (callable): The method that actually dispatches each RPC.
            This method will always be invoked inside of the event loop.
        loop (BackgroundEventLoop): The underlying event loop to use.
    """

    def __init__(self, rpc_handler, loop=SharedLoop):
        self._loop = loop
        self._thread = None
        self._started = False
        self._tasks = {}
        self._rpc_queue = RPCQueue(self._loop.get_loop(), rpc_handler)
        self._work_queues = set([self._rpc_queue])
        self._events = set()
        self._logger = logging.getLogger(__name__)

    def create_event(self, register=False):
        """Create an asyncio.Event inside the emulation loop.

        This method exists as a convenience to create an Event object that is
        associated with the correct EventLoop().  If you pass register=True,
        then the event will be registered as an event that must be set for the
        EmulationLoop to be considered idle.  This means that whenever
        wait_idle() is called, it will block until this event is set.

        Examples of when you may want this behavior is when the event is
        signaling whether a tile has completed restarting itself.  The reset()
        rpc cannot block until the tile has initialized since it may need to
        send its own rpcs as part of the initialization process.  However, we
        want to retain the behavior that once the reset() rpc returns the tile
        has been completely reset.

        The cleanest way of achieving this is to have the tile set its
        self.initialized Event when it has finished rebooting and register
        that event so that wait_idle() nicely blocks until the reset process
        is complete.

        Args:
            register (bool): Whether to register the event so that wait_idle
                blocks until it is set.

        Returns:
            asyncio.Event: The Event object.
        """

        event = self._loop.create_event()
        if register:
            self._events.add(event)

        return event

    def create_queue(self, register=False):
        """Create a new work queue and optionally register it.

        This will make sure the queue is attached to the correct event loop.
        You can optionally choose to automatically register it so that
        wait_idle() will block until the queue is empty.

        Args:
            register (bool): Whether to call register_workqueue() automatically.

        Returns:
            asyncio.Queue: The newly created queue.
        """

        queue = self._loop.create_queue()
        if register:
            self._work_queues.add(queue)

        return queue

    def get_current_rpc(self):
        """Get the currently running RPC for asynchronous responses.

        Returns the address and rpc_id of the RPC that is currently being
        dispatched. This information can be saved and passed later to
        finish_async_rpc() to send a response to an asynchronous rpc.

        Returns:
            (address, rpc_id): A tuple with the currently running RPC.
        """

        self.verify_calling_thread(True, "Only the emulation thread is allowed to inspect the current rpc")
        return self._rpc_queue.get_current_rpc()

    def finish_async_rpc(self, address, rpc_id, *response):
        """Finish a previous asynchronous RPC.

        This method should be called by a peripheral tile that previously
        had an RPC called on it and chose to response asynchronously by
        raising ``AsynchronousRPCResponse`` in the RPC handler itself.

        The response passed to this function will be returned to the caller
        as if the RPC had returned it immediately.

        This method must only ever be called from a coroutine inside the
        emulation loop that is handling background work on behalf of a tile.

        Args:
            address (int): The tile address the RPC was called on.
            rpc_id (int): The ID of the RPC that was called.
            *response: The response that should be returned to the caller.

                This can either be a single bytes or bytearray object or a
                str object containing the format code followed by the required
                number of python objects that will then be packed using
                pack_rpc_payload(format, args).

                If you pass no additional response arguments then an
                empty response will be given.
        """

        self.verify_calling_thread(True, "All asynchronous rpcs must be finished from within the emulation loop")

        if len(response) == 0:
            response_bytes = b''
        elif len(response) == 1:
            response_bytes = response[0]

            if not isinstance(response_bytes, (bytes, bytearray)):
                raise ArgumentError("When passing a binary response to finish_async_rpc, you must "
                                    "pass a bytes or bytearray object", response=response_bytes)
        else:
            resp_format = response[0]
            resp_args = response[1:]

            if not isinstance(resp_format, str):
                raise ArgumentError("When passing a formatted response to finish_async_rpc, you must "
                                    "pass a str object with the format code as the first parameter after "
                                    "the rpc id.", resp_format=resp_format, additional_args=resp_args)

            response_bytes = pack_rpc_payload(resp_format, resp_args)

        self._rpc_queue.finish_async_rpc(address, rpc_id, response_bytes)

    def start(self):
        """Start the background emulation loop."""

        if self._started is True:
            raise ArgumentError("EmulationLoop.start() called multiple times")

        self._loop.start()
        self._rpc_queue.start()
        self._started = True

    def stop(self):
        """Stop the background emulation loop."""

        if self._started is False:
            raise ArgumentError("EmulationLoop.stop() called without calling start()")

        self.verify_calling_thread(False, "Cannot call EmulationLoop.stop() from inside the event loop")
        self._loop.run_coroutine(self._clean_shutdown())

    def wait_idle(self, timeout=1.0):
        """Wait until the rpc queue is empty.

        This method may be called either from within the event loop or from
        outside of it.  If it is called outside of the event loop it will
        block the calling thread until the rpc queue is temporarily empty.

        If it is called from within the event loop it will return an awaitable
        object that can be used to wait for the same condition.

        The awaitable object will already have a timeout if the timeout
        parameter is passed.

        Args:
            timeout (float): The maximum number of seconds to wait.
        """

        async def _awaiter():
            background_work = {x.join() for x in self._work_queues}
            for event in self._events:
                if not event.is_set():
                    background_work.add(event.wait())

            _done, pending = await asyncio.wait(background_work, timeout=timeout)
            if len(pending) > 0:
                raise TimeoutExpiredError("Timeout waiting for event loop to become idle", pending=pending)

        if self.on_emulation_thread():
            return asyncio.wait_for(_awaiter(), timeout=timeout)

        self.run_task_external(_awaiter())
        return None

    def run_task_external(self, coroutine):
        """Inject a task into the emulation loop and wait for it to finish.

        The coroutine parameter is run as a Task inside the EmulationLoop
        until it completes and the return value (or any raised Exception) is
        pased back into the caller's thread.

        Args:
            coroutine (coroutine): The task to inject into the event loop.

        Returns:
            object: Whatever the coroutine returned.
        """

        self.verify_calling_thread(False, 'run_task_external must not be called from the emulation thread')
        return self._loop.run_coroutine(coroutine)

    def call_rpc_external(self, address, rpc_id, arg_payload, timeout=10.0):
        """Call an RPC from outside of the event loop and block until it finishes.

        This is the main method by which a caller outside of the EmulationLoop
        can inject an RPC into the EmulationLoop and wait for it to complete.
        This method is synchronous so it blocks until the RPC completes or the
        timeout expires.

        Args:
            address (int): The address of the mock tile this RPC is for
            rpc_id (int): The number of the RPC
            payload (bytes): A byte string of payload parameters up to 20 bytes
            timeout (float): The maximum time to wait for the RPC to finish.

        Returns:
            bytes: The response payload from the RPC
        """

        self.verify_calling_thread(False, "call_rpc_external is for use **outside** of the event loop")

        response = CrossThreadResponse()

        self._loop.get_loop().call_soon_threadsafe(self._rpc_queue.put_rpc, address, rpc_id, arg_payload, response)

        try:
            return response.wait(timeout)
        except RPCRuntimeError as err:
            return err.binary_error

    async def call_rpc_internal(self, address, rpc_id, arg_payload, timeout=10.0):
        """Call an RPC from inside of the event loop and yield until it finishes.

        This is the main method by which a caller inside of the EmulationLoop
        can inject an RPC into the EmulationLoop and wait for it to complete.

        Args:
            address (int): The address of the mock tile this RPC is for
            rpc_id (int): The number of the RPC
            payload (bytes): A byte string of payload parameters up to 20 bytes
            timeout (float): The maximum time to wait for the RPC to finish.

        Returns:
            bytes: The response payload from the RPC
        """

        self.verify_calling_thread(True, "call_rpc_internal is for use **inside** of the event loop")

        response = AwaitableResponse()
        self._rpc_queue.put_rpc(address, rpc_id, arg_payload, response)

        try:
            resp_payload = await response.wait(timeout)
        except RPCRuntimeError as err:
            resp_payload = err.binary_error

        return resp_payload

    async def await_rpc(self, address, rpc_id, *args, **kwargs):
        """Send an RPC from inside the EmulationLoop.

        This is the primary method by which tasks running inside the
        EmulationLoop dispatch RPCs.  The RPC is added to the queue of waiting
        RPCs to be drained by the RPC dispatch task and this coroutine will
        block until it finishes.

        **This method must only be called from inside the EmulationLoop**

        Args:
            address (int): The address of the tile that has the RPC.
            rpc_id (int): The 16-bit id of the rpc we want to call
            *args: Any required arguments for the RPC as python objects.
            **kwargs: Only two keyword arguments are supported:
                - arg_format: A format specifier for the argument list
                - result_format: A format specifier for the result

        Returns:
            list: A list of the decoded response members from the RPC.
        """

        self.verify_calling_thread(True, "await_rpc must be called from **inside** the event loop")

        if isinstance(rpc_id, RPCDeclaration):
            arg_format = rpc_id.arg_format
            resp_format = rpc_id.resp_format
            rpc_id = rpc_id.rpc_id
        else:
            arg_format = kwargs.get('arg_format', None)
            resp_format = kwargs.get('resp_format', None)

        arg_payload = b''

        if arg_format is not None:
            arg_payload = pack_rpc_payload(arg_format, args)

        self._logger.debug("Sending rpc to %d:%04X, payload=%s", address, rpc_id, args)

        response = AwaitableResponse()
        self._rpc_queue.put_rpc(address, rpc_id, arg_payload, response)

        try:
            resp_payload = await response.wait(1.0)
        except RPCRuntimeError as err:
            resp_payload = err.binary_error

        if resp_format is None:
            return []

        resp = unpack_rpc_payload(resp_format, resp_payload)
        return resp

    def verify_calling_thread(self, should_be_emulation, message=None):
        """Verify if the calling thread is or is not the emulation thread.

        This method can be called to make sure that an action is being taken
        in the appropriate context such as not blocking the event loop thread
        or modifying an emulate state outside of the event loop thread.

        If the verification fails an InternalError exception is raised,
        allowing this method to be used to protect other methods from being
        called in a context that could deadlock or cause race conditions.

        Args:
            should_be_emulation (bool): True if this call should be taking place
                on the emulation, thread, False if it must not take place on
                the emulation thread.
            message (str): Optional message to include when raising the exception.
                Otherwise a generic message is used.

        Raises:
            InternalError: When called from the wrong thread.
        """

        if should_be_emulation == self.on_emulation_thread():
            return

        if message is None:
            message = "Operation performed on invalid thread"

        raise InternalError(message)

    def add_task(self, tile_address, coroutine):
        """Add a task into the event loop.

        This is the main entry point for registering background tasks that are
        associated with a tile. The tasks are added to the EmulationLoop and
        the tile they are a part of is recorded.  When the tile is reset, all
        of its background tasks are canceled as part of the reset process.

        If you have a task that should not be associated with any tile, you
        may pass `None` for tile_address and the task will not be cancelled
        when any tile is reset.

        Args:
            tile_address (int): The address of the tile running
                the task.
            coroutine (coroutine): A coroutine that will be added
                to the event loop.
        """

        self._loop.get_loop().call_soon_threadsafe(self._add_task, tile_address, coroutine)

    def is_tile_busy(self, address):
        """Check that tile has no pending rpcs

        Args:
            address (int): The address of the time

        Returns:
            bool: True if there is a rpc for the tile already in the queue
        """
        return self._rpc_queue.is_pending_rpc(address)

    def on_emulation_thread(self):
        """Returns whether we are running on the emulation thread.

        Returns:
            bool: True if we are on the emulation thread, else False.
        """

        return self._loop.inside_loop()

    async def stop_tasks(self, address):
        """Clear all tasks pertaining to a tile.

        This coroutine will synchronously cancel all running tasks that were
        attached to the given tile and wait for them to stop before returning.

        Args:
            address (int): The address of the tile we should stop.
        """

        tasks = self._tasks.get(address, [])
        for task in tasks:
            task.cancel()

        asyncio.gather(*tasks, return_exceptions=True)
        self._tasks[address] = []

    async def _clean_shutdown(self):
        """Cleanly shutdown the emulation loop."""

        # Cleanly stop any other outstanding tasks not associated with tiles
        remaining_tasks = []
        for task in self._tasks.get(None, []):
            self._logger.debug("Cancelling task at shutdown %s", task)
            task.cancel()

            remaining_tasks.append(task)

        asyncio.gather(*remaining_tasks, return_exceptions=True)

        if len(remaining_tasks) > 0:
            del self._tasks[None]

        # Shutdown tasks associated with each tile
        remaining_tasks = []

        for address in sorted(self._tasks, reverse=True):
            if address is None:
                continue

            self._logger.debug("Shutting down tasks for tile at %d", address)
            for task in self._tasks.get(address, []):
                task.cancel()
                remaining_tasks.append(task)

        asyncio.gather(*remaining_tasks, return_exceptions=True)

        await self._rpc_queue.stop()

    def _add_task(self, tile_address, coroutine):
        """Add a task from within the event loop.

        All tasks are associated with a tile so that they can be cleanly
        stopped when that tile is reset.
        """

        self.verify_calling_thread(True, "_add_task is not thread safe")

        if tile_address not in self._tasks:
            self._tasks[tile_address] = []

        task = self._loop.get_loop().create_task(coroutine)
        self._tasks[tile_address].append(task)



"""Declarations of all important constants related to physical IOTile devices."""

from inspect import getmembers
from iotile.sg import DataStream
from .errors import Error, ConfigDatabaseError, SensorLogError, SensorGraphError
from . import rpcs
from . import streams
from .const_tilemanager import RunLevel, TileState
from .const_subsystems import ControllerSubsystem

_RPC_NAME_MAP = {rpc_decl.rpc_id: name for name, rpc_decl in
                 getmembers(rpcs, lambda x: isinstance(x, rpcs.RPCDeclaration))}

_STREAM_NAME_MAP = {stream_decl: name for name, stream_decl in
                    getmembers(streams, lambda x: isinstance(x, int))}


def rpc_name(rpc_id):
    """Map an RPC id to a string name.

    This function looks the RPC up in a map of all globally declared RPCs,
    and returns a nice name string.  if the RPC is not found in the global
    name map, returns a generic name string such as 'rpc 0x%04X'.

    Args:
        rpc_id (int): The id of the RPC that we wish to look up.

    Returns:
        str: The nice name of the RPC.
    """

    name = _RPC_NAME_MAP.get(rpc_id)
    if name is None:
        name = 'RPC 0x%04X' % rpc_id

    return name


def stream_name(stream_id):
    """Map a stream id to a human readable name.

    The mapping process is as follows:

    If the stream id is globally known, its global name is used as <name>
    otherwise a string representation of the stream is used as <name>.

    In both cases the hex representation of the stream id is appended as a
    number:

    <name> (0x<stream id in hex>)

    Args:
        stream_id (int): An integer stream id.

    Returns:
        str: The nice name of the stream.
    """

    name = _STREAM_NAME_MAP.get(stream_id)
    if name is None:
        name = str(DataStream.FromEncoded(stream_id))

    return "{} (0x{:04X})".format(name, stream_id)


def pack_error(subsystem, code):
    """Pack a 32-bit error code from a subsystem id and a short error code.

    Args:
        subsystem (int): The subsystem declaring the error.
        code (int): The 16-bit error code.

    Returns:
        int: The 32-bit packed error code.
    """

    return ((subsystem & 0xFFFF) << 16) | (code & 0xFFFF)


def unpack_error(error):
    """Unpack a 32-bit error code into a subsystem and short code.

    In the future this routine will attempt to match the error code and
    subsystem id with a IntEnum subclass to get nice nices when printed.

    Args:
        error (int): A packed 32-bit error code.

    Returns:
        (int, int): The subsystem ID and short error code.
    """

    return (error >> 16), (error & 0xFFFF)


__all__ = ['Error', 'ConfigDatabaseError', 'SensorLogError', 'streams',
           'stream_name', 'rpcs', 'rpc_name', 'pack_error', 'unpack_error', 'RunLevel',
           'TileState', 'ControllerSubsystem']



"""RPCs implemented by the RawSensorLog (RSL) subsystem on an IOTile controller.

The raw sensor log, abbreviated RSL, is the persistent storage subsystem
inside of an IOTile controller.  It's job is to provide a series of named
FIFOs that can be used as multi-producer, multi-consumer queues.  Each queue
is assigned a name and any consumer that attaches to it can poll for any new
data since the last time they checked.

Each queue is assigned a storage class which specifies whether the data in the
queue is persistent, so it will survice a device reset or volatile, so it will
not.  Volatile queues are also called "virtual" queues inside the RSL, since
they do not need to actually allocate a persistent FIFO data structure while
still giving the appearance that one exists.

The major subsystems that interact with the RSL are SensorGraph and
StreamManager.

  - SensorGraph uses a series of rules to produce data and assigns each data a
    "stream" name.  These streams map 1-1 onto the RSL queues, so there is a
    single RSL queue for every SensorGraph stream.  The terms queue, FIFO and
    stream in this context are 100% synonomous.

    Internally Sensorgraph also uses streams to cache data that will be
    consumed by later rules.

  - StreamManager consumes streams/queues from the RSL and when given
    conditions are met, it packages up all or a subset of the readings in a
    given set of queues into a report, which is a a signed list of readings
    that can be sent out of the IOTile device through its Streaming Interface.

In most subsequent documentation, the word **stream** will be used to refer to
uniformly to both virtual and persistent queues allocated in the RSL
subsystem.

ALl errors returned by an RPC declared below will have the subsystem set to
SENSOR_LOG_SUBSYSTEM and the error code set to something in Error or
SensorLogError.

The RPCs below allow for direct inspection and interaction with the RSL. In
particular, they can be used for:

  - pushing readings directly to the RSL subsystem
  - querying status on memory usage
  - dumping all readings by stream name
  - clearing all data from the RSL
"""

from iotile.core.hw.virtual import RPCDeclaration


RSL_PUSH_READING = RPCDeclaration(0x2000, "LH", "L")
"""Push a single reading into a stream.

You call this RPC with the 32-bit value you want to push and the 16-bit stream
name and this RPC synchronously pushes the value to the stream.  This is a
direct push to the RSL that the does not go through the SensorGraph subsystem
at all so it is not guaranteed to immediately trigger any sensor graph rules
that may apply to that stream unless the stream is of class input.

Args:
  - uint32_t: The new value that we want to push.  This can be anything that
    fits in 32-bits.
  - uin16_t: An encoded StreamID that specifies the stream that we wish to
    push to.

Returns:
  - uint32_t: An error Code.

Possible Errors:
  - RING_BUFFER_FULL: The persistent RSL queue that would store this reading
    is in fill-stop mode and it is currently full.
  - VIRTUAL_STREAM_NOT_FOUND: The stream name passed specifies a virtual
    stream and there are no stream walkers configured that listen to that stream.
    Virtual streams directly store their readings in the walkers of all of their
    consumers so there is no place for this reading to go.
"""

RSL_PUSH_MANY_READINGS = RPCDeclaration(0x2013, "LLH", "LL")
"""Push multiple copies of the same value to a given stream.

This RPC is primarily useful for testing purposes.  It can quickly fill the
RSL with data.  Conceptually this is equivalent to N calls to RSL_PUSH_READING
with the same value and stream id but since it happens inside a single RPC, it
is signficantly faster (no RPC overhead per reading).

In practice this RPC be 100-1000 times faster when you need to push a lot of
data and don't need the values to be different.

This RPC stops as soon as the first error is returned from pushing a reading.
You can learn how many readings were actually pushed by inspecting the return
value.

Args:
  - uint32_t: The value to push N copies of
  - uint32_t: The number of copies to push.
  - uint16_t: The stream to push the readings to.

Returns:
  - uint32_t: An error code.
  - uint32_t: The number of readings that were successfully pushed.

Possible Errors:
  - RING_BUFFER_FULL: The persistent RSL queue that would store this reading
    is in fill-stop mode and it is currently full.
  - VIRTUAL_STREAM_NOT_FOUND: The stream name passed specifies a virtual
    stream and there are no stream walkers configured that listen to that stream.
    Virtual streams directly store their readings in the walkers of all of their
    consumers so there is no place for this reading to go.
"""

RSL_COUNT_READINGS = RPCDeclaration(0x2001, "", "3L")
"""Count how many persistent readings are stored in the RSL.

The RSL maintains two persistent storage areas: one for output class streams
and one for storage (also called buffered) class streams.  This RPC just
returns the count of readings stored in each of those two persistent storage
areas.

Returns:
  - uint32_t: An error code.  No error codes are currently possible.
  - uint32_t: The number of readings currently stored in the storage area of the
    RSL.
  - uint32_t: The number of readings currently stored in the output area of the
    RSL.
"""

RSL_HIGHEST_READING_ID = RPCDeclaration(0x2011, "", "LL")
"""Return the highest reading ID current in the RSL.

All readings in the RSL that are stored persistently are assigned a
monotonically increasing sequence number called their reading id.

This RPC returns the highest ID that is currently stored in the RSL. These
unique IDs are also assigned to non-reading objects like streamer reports so
the return value of this RPC is not guaranteed to be the highest ID ever
allocated but it is guaranteed to be the highest ID ever allocated to a
reading.

Returns:
  - uint32_t: An error code.  No errors are currently possible.
  - uint32_t: The highest reading ID currently assigned to a reading in the
    RSL.
"""

RSL_CLEAR_READINGS = RPCDeclaration(0x0200c, "", "L")
"""Clear all of the readings inside the RSL.

This will clear both persistently stored readings as well as all virtual
readings except for constant class streams.  This method pushes a single
system output reading to indicate that the RSL has been cleared so the result
of this RPC will be:
  - all virtual streams cleared except constant streams
  - all current storage and output streams cleared
  - a single output reading pushed

Returns:
  - uint32_t: An error code.  There are currently no possible errors.
"""

RSL_INSPECT_VIRTUAL_STREAM = RPCDeclaration(0x200b, "H", "LL")
"""Inspect the value current stored in a virtual stream.

This RPC will show the value inside of a virtual stream.  Since virtual
streams do not have any persistent storage backing, they only have room
for a single value, which this RPC returns.

Also, since virtual streams store their values directly in the stream
walkers that are consuming their data, if there are no stream walkers
allocated for a given virtual stream, this RPC will return an error
since there is no data location to inspect.

Args:
  - uint16_t: The stream ID that we wish to inspect.  This should be
    a virtual class stream.

Returns:
  - uint32_t: An error code.
  - uint32_t: The value currently stored in the named stream.  This is
    only valid if NO_ERROR is returned.

Possible Errors:
  - VIRTUAL_STREAM_NOT_FOUND: There is no allocated stream walker that
    listens on the given virtual stream.
"""

RSL_DUMP_STREAM_BEGIN = RPCDeclaration(0x2008, "H", "LLLL")
"""Start dumping all data in a given persistent stream.

Unlike most RPCs, this RPC is not atomic and must be called in a proper
sequence for correct results.  The purpose of the sequence of RPCs is to
download all of the readings stored in a given stream in the RSL.  Since
there is a variable amount of data in the stream, the sequence of RPCs
needed to download it are:

  - RSL_DUMP_STREAM_BEGIN: Start the dump process and allocate an internal
    controller structure to track where we are in the stream.
  - RSL_DUMP_STREAM_SEEK (Optional): Optionally seek to a specific point in
    the stream by reading id.  If this RPC is not sent, the dump always starts
    at the beginning of the stream.
  - RSL_DUMP_STREAM_NEXT: Return the next reading in the stream and an error
    if we are at the end of the stream.

It is always safe to call RSL_DUMP_STREAM_BEGIN but this will clear out the
state from any previous dumping process so the client must ensure that only a
single dump happens at a time.

Args:
  - uint16_t: The encoded stream id of the stream that we wish to dump.  This
    must be a storage or output class stream.  If you are trying to inspect a
    virtual stream, use RSL_INSPECT_VIRTUAL_STREAM instead.

Returns:
  - uint32_t: An error code covering the allocation of the internal walker
    resource. No current error codes are possible.
  - uint32_t: A second error code covering the counting of available readings
    in the stream.  No current error codes are possible.
  - uint32_t: The number of available readings.  You should be able to call
    RSL_DUMP_STREAM_NEXT exactly this many times without error.  However, keep
    in mind that readings may be added or overwritten as you iterate so you
    should inspect the actual return value of RSL_DUMP_STREAM_NEXT to know
    when you are finished.
  - uint32_t: The current uptime of the controller.  This is necessary for
    proper conversion of readings stamped with controller uptime into actual
    UTC time.
"""

RSL_DUMP_STREAM_SEEK = RPCDeclaration(0x2012, "L", "3L")
"""Seek to a specific reading id inside the stream we are currently dumping.

See RSL_DUMP_STREAM_BEGIN for more details on the stream dumping process.
This method is optional and will move the read pointer so that it points to
the reading with the given reading ID.

Make sure you understand the behavior of what happens when you seek to a
stream id that either does not exist or is not part of the stream in question.

There are 3 possible cases that can happen when seeking a stream by reading_id:

  - The reading ID could not exist.  In this case the current location of the
    stream is unchanged and an error is returned.
  - The reading ID could be found but correspond with a reading not in this
    stream. In this case the stream location is set to the first reading with
    id after the desired seek_id.  An error code is returned but the stream
    location is changed.
  - The reading ID is found within the given stream.  In this case the stream
    location is updated to point at the specific reading id and NO_ERROR is
    returned.

Args:
  - uint32_t: The reading ID to seek.

Returns:
  - uint32_t: An error code covering the seek process.
  - uint32_t: An error code covering the counting of available readings.
  - uint32_t: The number of available readings after the seek.

Possible Errors:
  - STREAM_WALKER_NOT_INITIALIZED: If this RPC was not called after first
    calling RSL_DUMP_STREAM_BEGIN to set the desired stream id and allocate
    internal resources.
  - CANNOT_USE_UNBUFFERED_STREAM: If the stream id passed to
    RSL_DUMP_STREAM_BEGIN points to a virtual stream.  Virtual streams do not
    have reading ids and cannot be seeked.
  - ID_FOUND_FOR_ANOTHER_STREAM: If the reading id passed was found but
    did not correspond to a reading inside the desired stream.
  - NO_MORE_READINGS: If the reading id was not found at all in the RSL.
"""

RSL_DUMP_STREAM_NEXT = RPCDeclaration(0x2009, "B", "V")
"""Dump the next reading in the previously selected stream.

See RSL_DUMP_STREAM_BEGIN for more details on the stream dumping process.

This RPC will dump a single reading from the currenly selected stream that was
passed to the last call to RSL_DUMP_STREAM_BEGIN.  It's signature has changed
so it takes a single byte argument value that indicates whether the caller
would prefer the new or old return format.  The new return format contains
more information and should always be chosen.  If this flag is omitted or
passed with the value of 0, the old return format will be used.

For the historical record, the old return format, that is no longer used is:
  - uint32_t: An error code.
  - uint32_t: The reading timestamp
  - uint32_t: The reading value

In the v0 return format, there was no way to know the actual stream id of the
reading or its unique reading id.

Args:
  - uint8_t: The desired return format.  If set to 0, an old format will be
    used otherwise, if 1 is passed, the new format will be used.  No value
    other than 0 or 1 should be passed.

Returns (format 0):
  - uint32_t: An error code.
  - uint32_t: The reading timestamp.
  - uint32_t: The reading value.

Returns (format 1):
  - uint32_t: An error code.
  - uint32_t: The reading timestamp.
  - uint32_t: The reading value.
  - uint32_t: The reading's unique reading ID.
  - uint16_t: The reading's stream ID.
  - uint16_t: Reserved, should be ignored.

Possible Errors:
  - NO_MORE_READINGS: If all of the available readings have been exhausted and
    there is no more data.
"""



"""Tests of the clock manager subsystem."""

import datetime
import pytest

from iotile.core.hw import HardwareManager
from iotile.core.exceptions import HardwareError
from iotile.emulate.reference import ReferenceDevice
from iotile.emulate.transport import EmulatedDeviceAdapter


@pytest.fixture(scope="function")
def basic_device():
    """A preprogrammed basic sensorgraph for testing."""

    device = ReferenceDevice({'simulate_time': False})
    adapter = EmulatedDeviceAdapter(None, devices=[device])

    nodes = [
        "(system input 2 always) => output 1 using copy_latest_a",
        "(system input 3 always) => output 2 using copy_latest_a",
        "(system input 5 always) => output 3 using copy_latest_a",
        "(system input 6 always) => output 4 using copy_latest_a"
    ]

    with HardwareManager(adapter=adapter) as hw:
        hw.connect(1)

        con = hw.get(8)
        sensor_graph = con.sensor_graph()

        for node in nodes:
            sensor_graph.add_node(node)

        sensor_graph.add_streamer('output 10', 'controller', True, 'individual', 'telegram')

        yield hw, device


def test_user_ticks(basic_device):
    """Make sure that we can control user ticks."""

    hw, device = basic_device

    con = hw.get(8, basic=False)
    sensor_graph = con.sensor_graph()
    config = con.config_database()

    assert sensor_graph.user_tick(0) == 0
    assert sensor_graph.user_tick(1) == 0
    assert sensor_graph.user_tick(2) == 0

    with pytest.raises(HardwareError):
        sensor_graph.user_tick(3)

    sensor_graph.set_user_tick(0, 1)
    sensor_graph.set_user_tick(1, 5)
    sensor_graph.set_user_tick(2, 6)

    with pytest.raises(HardwareError):
        sensor_graph.set_user_tick(3, 1)

    assert sensor_graph.user_tick(0) == 1
    assert sensor_graph.user_tick(1) == 5
    assert sensor_graph.user_tick(2) == 6

    # Make sure set on reset works
    config.set_variable('controller', 0x2000, 'uint32_t', 2)
    config.set_variable('controller', 0x2002, 'uint32_t', 4)
    config.set_variable('controller', 0x2003, 'uint32_t', 7)

    con.reset(wait=0)
    device.wait_idle()

    assert sensor_graph.user_tick(0) == 2
    assert sensor_graph.user_tick(1) == 4
    assert sensor_graph.user_tick(2) == 7


def test_tick_inputs(basic_device):
    """Test to make sure that ticks are sent to sensor_graph."""

    hw, device = basic_device

    con = hw.get(8, basic=False)
    sensor_graph = con.sensor_graph()
    # sensor_graph = find_proxy_plugin('iotile_standard_library/lib_controller', 'SensorGraphPlugin')(con)
    clock_man = device.controller.clock_manager

    sensor_graph.enable()

    assert clock_man.uptime == 0

    sensor_graph.set_user_tick(0, 1)
    sensor_graph.set_user_tick(1, 2)
    sensor_graph.set_user_tick(2, 5)

    for i in range(1, 11):
        device.controller.clock_manager.handle_tick()
        assert clock_man.uptime == i

    dump1 = sensor_graph.download_stream('output 1')
    assert len(dump1) == 1
    key_parts_1 = [(x.raw_time, x.value) for x in dump1]
    assert key_parts_1 == [(10, 10)]

    dump2 = sensor_graph.download_stream('output 2')
    assert len(dump2) == 10
    key_parts_2 = [(x.raw_time, x.value) for x in dump2]
    assert key_parts_2 == list(zip(range(1, 11), range(1, 11)))

    dump3 = sensor_graph.download_stream('output 3')
    assert len(dump3) == 5
    key_parts_3 = [(x.raw_time, x.value) for x in dump3]
    assert key_parts_3 == list(zip(range(2, 11, 2), range(2, 11, 2)))

    dump4 = sensor_graph.download_stream('output 4')
    assert len(dump4) == 2
    key_parts_4 = [(x.raw_time, x.value) for x in dump4]
    assert key_parts_4 == list(zip(range(5, 11, 5), range(5, 11, 5)))


def test_utc_time(basic_device):
    """Make sure we can get and set utc time."""

    hw, device = basic_device

    con = hw.get(8, basic=False)
    test_interface = con.test_interface()

    test_time = datetime.datetime(2018, 11, 11, 16, 0, 0)
    zero = datetime.datetime(1970, 1, 1)
    y2k_zero = datetime.datetime(2000, 1, 1)

    device.controller.clock_manager.handle_tick()

    delta = (test_time - zero).total_seconds()
    y2k_delta = (test_time - y2k_zero).total_seconds()

    test_interface.synchronize_clock(delta)
    device_time = test_interface.current_time()
    device_uptime = test_interface.get_uptime()
    info = test_interface.get_timeoffset()

    assert device_time & (1 << 31)
    assert (device_time & ~(1 << 31)) == int(y2k_delta)
    assert device_uptime == 1
    assert info == {'is_utc': True, 'offset': int(y2k_delta) - 1}



"""All well-known globally defined RPCS including their RPC identifier and signatures.

All of the RPC declarations in this module are of type RPCDeclaration and include
a detailed description of their expected semantics in the associated docstring.

The RPCDeclaration can be directly passed to the EmulatedDevice.rpc or
EmulatedDevice.deferred_rpc functions as `rpc_id`, in which case arg_format
and resp_format keyword arguments can be omitted as they will be inferred from
the RPCDeclaration itself.
"""

#pylint:disable=wildcard-import,unused-wildcard-import;This is designed to a single place for all RPCS.
from iotile.core.hw.virtual import RPCDeclaration
from .rpc_tilemanager import *
from .rpc_config_database import *
from .rpc_config_variables import *
from .rpc_sensorlog import *
from .rpc_sensorgraph import *
from .rpc_clockmanager import *
from .rpc_controller import *


# Tile Lifecycle Related RPCS

RESET = RPCDeclaration(1, "", "")
"""Immediately reset the tile.

This RPC will never complete in the normal sense because the tile will reset
in the middle of executing it.   Callers should expect to receive a
ModuleNotFound error or other atypical response code from this RPC.  There is
not a straightforward, portable way to determine the response code caused by
the tile resetting from a response code caused by an issue running this RPC.
"""

HARDWARE_VERSION = RPCDeclaration(2, "", "10s")
"""Get a hardware identification string.

This RPC returns a string of up to 10 bytes (padded with null bytes) to
a fixed 10 byte length that contains a hardware identification string.
The meaning of the string is opaque and defined by the firmware that is
returning this string.  The typical use of this value is to determine
what tile hardware is running if a given tile firmware is compatible
with multiple boards.

Returns:
  - char[10]: A utf-8 hardware string padded with null bytes to 10
    bytes.
"""

TILE_STATUS = RPCDeclaration(4, "", "H6s3BB")
"""Query the tile's name and status.

This RPC must be implemented by all tiles and is primarily used to return the
6-character name string that identifies the firmware image running on the
tile. This string is matched against installed proxy classes to find the
appropriate proxy object to use for wrapping the tile.

It can also be used to query the firmware version of the tile or a set of flag
bits about the tile's current state.

Returns:
  - uint16_t: A numerical hardware identifier for the architecture the tile
    is running on.
  - char[6]: The tile firmware's 6-character module identifier that is used to
    match the tile with an appropriate proxy class.
  - uint8_t[3]: The tile's firmware version stored as major, minor, patch.
  - uint8_t: A set of bit flags defining the current state of the tile such as
    whether it is running, trapped in a panic routine, etc.
"""

START_APPLICATION = RPCDeclaration(6, "", "")
"""Inform the peripheral tile that it should pass control to its application firmware.

Once the controller has finished streaming all configuration information to a tile
it sends the tile this RPC so that it can check its configuration state and if
everything is valid pass control to application firmware.
"""



"""Base class for virtual tiles designed to emulate physical tiles."""

import struct
import base64
import logging
from iotile.core.exceptions import ArgumentError, DataError
from iotile.core.hw.exceptions import TileNotFoundError
from iotile.core.hw.virtual import tile_rpc, RPCDispatcher
from .emulated_device import EmulatedDevice
from .emulation_mixin import EmulationMixin
from ..constants import rpcs, Error


class ConfigDescriptor(object):
    """Helper class for declaring a configuration variable."""

    def __init__(self, config_id, type_name, default=None, name=None, python_type=None):
        self.config_id = config_id
        self.total_size, self.unit_size, self.base_type, self.variable = parse_size_name(type_name)
        self.max_units = self.total_size // self.unit_size
        self.type_name = type_name
        self.current_value = bytearray()
        self.special_type = python_type
        self.default_value = self._convert_default_value(default)

        self._validate_python_type(python_type)

        if name is None:
            name = "Unnamed variable 0x%X" % config_id

        self.name = name

    def _validate_python_type(self, python_type):
        """Validate the possible combinations of python_type and type_name."""

        if python_type == 'bool':
            if self.variable:
                raise ArgumentError("You can only specify a bool python type on a scalar (non-array) type_name", type_name=self.type_name)

            return

        if python_type == 'string':
            if not (self.variable and self.unit_size == 1):
                raise ArgumentError("You can only pass a string python type on an array of 1-byte objects", type_name=self.type_name)

            return

        if python_type is not None:
            raise ArgumentError("You can only declare a bool or string python type.  Otherwise it must be passed as None", python_type=python_type)

    def _convert_default_value(self, default):
        """Convert the passed default value to binary.

        The default value (if passed) may be specified as either a `bytes`
        object or a python int or list of ints.  If an int or list of ints is
        passed, it is converted to binary.  Otherwise, the raw binary data is
        used.

        If you pass a bytes object with python_type as True, do not null terminate
        it, an additional null termination will be added.

        Passing a unicode string is only allowed if as_string is True and it
        will be encoded as utf-8 and null terminated for use as a default value.
        """

        if default is None:
            return None

        if isinstance(default, str):
            if self.special_type == 'string':
                return default.encode('utf-8') + b'\0'

            raise DataError("You can only pass a unicode string if you are declaring a string type config variable", default=default)

        if isinstance(default, (bytes, bytearray)):
            if self.special_type == 'string' and isinstance(default, bytes):
                default += b'\0'

            return default

        if isinstance(default, int):
            default = [default]

        format_string = "<" + (self.base_type*len(default))
        return struct.pack(format_string, *default)

    def clear(self):
        """Clear this config variable to its reset value."""

        if self.default_value is None:
            self.current_value = bytearray()
        else:
            self.current_value = bytearray(self.default_value)

    def update_value(self, offset, value):
        """Update the binary value currently stored for this config value.

        Returns:
            int: An opaque error code that can be returned from a set_config rpc
        """

        if offset + len(value) > self.total_size:
            return Error.INPUT_BUFFER_TOO_LONG

        if len(self.current_value) < offset:
            self.current_value += bytearray(offset - len(self.current_value))
        if len(self.current_value) > offset:
            self.current_value = self.current_value[:offset]

        self.current_value += bytearray(value)
        return 0

    def latch(self):
        """Convert the current value inside this config descriptor to a python object.

        The conversion proceeds by mapping the given type name to a native
        python class and performing the conversion.  You can override what
        python object is used as the destination class by passing a
        python_type parameter to __init__.

        The default mapping is:
        - char (u)int8_t, (u)int16_t, (u)int32_t: int
        - char[] (u)int8_t[], (u)int16_t[]0, u(int32_t): list of int

        If you want to parse a char[] or uint8_t[] as a python string, it
        needs to be null terminated and you should pass python_type='string'.

        If you are declaring a scalar integer type and wish it to be decoded
        as a bool, you can pass python_type='bool' to the constructor.

        All integers are decoded as little-endian.

        Returns:
            object: The corresponding python object.

            This will either be an int, list of int or string based on the
            type_name specified and the optional python_type keyword argument
            to the constructor.

        Raises:
            DataError: if the object cannot be converted to the desired type.
            ArgumentError: if an invalid python_type was specified during construction.
        """

        if len(self.current_value) == 0:
            raise DataError("There was no data in a config variable during latching", name=self.name)

        # Make sure the data ends on a unit boundary.  This would have happened automatically
        # in an actual device by the C runtime 0 padding out the storage area.
        remaining = len(self.current_value) % self.unit_size
        if remaining > 0:
            self.current_value += bytearray(remaining)

        if self.special_type == 'string':
            if self.current_value[-1] != 0:
                raise DataError("String type was specified by data did not end with a null byte", data=self.current_value, name=self.name)

            return bytes(self.current_value[:-1]).decode('utf-8')

        fmt_code = "<" + (self.base_type * (len(self.current_value) // self.unit_size))
        data = struct.unpack(fmt_code, self.current_value)

        if self.variable:
            data = list(data)
        else:
            data = data[0]

            if self.special_type == 'bool':
                data = bool(data)

        return data


#pylint:disable=abstract-method;This is an abstract base class
class EmulatedTile(EmulationMixin, RPCDispatcher):
    """Base class for virtual tiles designed to emulate physical tiles.

    This class adds additional state and test scenario loading functionality
    as well as tracing of state changes on the emulated device for comparison
    and verification purposes.

    There is a small set of behavior that all tiles must implement which is
    implemented in this class including a base set of RPCs for setting
    config variables and getting the tile's name.

    Args:
        address (int): The address of this tile in the VirtualIOTIleDevice
            that contains it
        device (TileBasedVirtualDevice): Device on which this tile is running.
            This parameter is not optional on EmulatedTiles.
    """

    __NO_EXTENSION__ = True

    hardware_type = 0
    """The hardware type is a single uint8_t that can be used to record the chip architecture used."""

    firmware_version = (1, 0, 0)
    """The version of the application firmware running on the tile."""

    executive_version = (1, 0, 0)
    """The version of the executive kernel running on the tile."""

    api_version = (3, 0)
    """The API version agreed between the application and executive kernel."""

    app_started = None
    """Default implementation of tiles does not have a separate phase before application code runs."""

    hardware_string = b'pythontile'
    """A 10 bytes identifier for the hardware that the tile is running on."""

    name = b'noname'
    """A 6-byte identifier for the application firmware running on the tile."""

    def __init__(self, address, device):
        if not isinstance(device, EmulatedDevice):
            raise ArgumentError("You can only add an EmulatedTile to an EmulatedDevice", device=device)

        RPCDispatcher.__init__(self)
        EmulationMixin.__init__(self, address, device.state_history)

        self._config_variables = {}
        self._device = device
        self._logger = logging.getLogger(__name__)
        self.address = address
        self.initialized = device.emulator.create_event(register=True)

    def declare_config_variable(self, name, config_id, type_name, default=None, convert=None):  #pylint:disable=too-many-arguments;These are all necessary with sane defaults.
        """Declare a config variable that this emulated tile accepts.

        The default value (if passed) may be specified as either a `bytes`
        object or a python int or list of ints.  If an int or list of ints is
        passed, it is converted to binary.  Otherwise, the raw binary data is
        used.

        Passing a unicode string is only allowed if as_string is True and it
        will be encoded as utf-8 and null terminated for use as a default value.

        Args:
            name (str): A user friendly name for this config variable so that it can
                be printed nicely.
            config_id (int): A 16-bit integer id number to identify the config variable.
            type_name (str): An encoded type name that will be parsed by parse_size_name()
            default (object): The default value if there is one.  This should be a
                python object that will be converted to binary according to the rules for
                the config variable type specified in type_name.
            convert (str): whether this variable should be converted to a
                python string or bool rather than an int or a list of ints.  You can
                pass either 'bool', 'string' or None
        """

        config = ConfigDescriptor(config_id, type_name, default, name=name, python_type=convert)
        self._config_variables[config_id] = config

    def reset_config_variables(self):
        """Clear the contents of all config variables to their defaults.

        This method should be used with caution.  It is designed to be called
        during the reset process of a tile in order to properly initialize its
        config variables.
        """

        for config in self._config_variables.values():
            config.clear()

    def latch_config_variables(self):
        """Latch the current value of all config variables as python objects.

        This function will capture the current value of all config variables
        at the time that this method is called.  It must be called after
        start() has been called so that any default values in the config
        variables have been properly set otherwise DataError will be thrown.

        Conceptually this method performs the operation that happens just
        before a tile executive hands control to the tile application
        firmware. It latches in the value of all config variables at that
        point in time.

        For convenience, this method does all necessary binary -> python
        native object conversion so that you just get python objects back.

        Returns:
            dict: A dict of str -> object with the config variable values.

            The keys in the dict will be the name passed to
            `declare_config_variable`.

            The values will be the python objects that result from calling
            latch() on each config variable.  Consult ConfigDescriptor.latch()
            for documentation on how that method works.
        """

        return {desc.name: desc.latch() for desc in self._config_variables.values()}

    def dump_state(self):
        """Dump the current state of this emulated tile as a dictionary.

        This function just dumps the status of the config variables.  It is designed to
        be called in a chained fashion to serialize the complete state of a tile subclass.

        Returns:
            dict: The current state of the object that could be passed to load_state.
        """

        return {
            "config_variables": {x: base64.b64encode(y.current_value).decode('utf-8') for x, y in self._config_variables.items()},
        }

    def restore_state(self, state):
        """Restore the current state of this emulated object.

        Args:
            state (dict): A previously dumped state produced by dump_state.
        """

        config_vars = state.get('config_variables', {})

        for str_name, str_value in config_vars.items():
            name = int(str_name)
            value = base64.b64decode(str_value)

            if name in self._config_variables:
                self._config_variables[name].current_value = value

    def start(self, channel=None):
        """Start any background workers on this tile."""
        pass

    def stop(self):
        """Stop any background workers on this tile."""
        pass

    def _handle_reset(self):
        """Hook to perform any required reset actions.

        Subclasses that override this method should always call
        super()._handle_reset() to make sure their base classes properly
        initialize their reset state.

        The only thing we need to properly do is make sure our config
        variables return to their reset states.
        """

        self.initialized.clear()
        self.reset_config_variables()

    async def _reset_vector(self):
        """Background task that is started when the tile is started.

        This method should be overriden by all tiles that have specific
        background functionality.  It must set the `self.initialized` event
        when it finishes initializing all of the tile's states.
        """

        self.initialized.set()

    async def reset(self):
        """Synchronously reset a tile.

        This method must be called from the emulation loop and will
        synchronously shut down all background tasks running this tile, clear
        it to reset state and then restart the initialization background task.
        """

        await self._device.emulator.stop_tasks(self.address)

        self._handle_reset()

        self._logger.info("Tile at address %d has reset itself.", self.address)

        self._logger.info("Starting main task for tile at address %d", self.address)
        self._device.emulator.add_task(self.address, self._reset_vector())

    @tile_rpc(*rpcs.RESET)
    async def reset_rpc(self):
        """Reset this tile."""
        await self.reset()
        raise TileNotFoundError("tile was reset via an RPC")

    @tile_rpc(*rpcs.LIST_CONFIG_VARIABLES)
    def list_config_variables(self, offset):
        """List defined config variables up to 9 at a time."""

        names = sorted(self._config_variables)
        names = names[offset:offset + 9]
        count = len(names)

        if len(names) < 9:
            names += [0]*(9 - count)

        return [count] + names

    @tile_rpc(*rpcs.DESCRIBE_CONFIG_VARIABLE)
    def describe_config_variable(self, config_id):
        """Describe the config variable by its id."""

        config = self._config_variables.get(config_id)
        if config is None:
            return [Error.INVALID_ARRAY_KEY, 0, 0, 0, 0]

        packed_size = config.total_size
        packed_size |= int(config.variable) << 15

        return [0, 0, 0, config_id, packed_size]

    @tile_rpc(*rpcs.SET_CONFIG_VARIABLE)
    def set_config_variable(self, config_id, offset, value):
        """Set a chunk of the current config value's value."""

        if self.initialized.is_set():
            return [Error.STATE_CHANGE_AT_INVALID_TIME]

        config = self._config_variables.get(config_id)
        if config is None:
            return [Error.INVALID_ARRAY_KEY]

        error = config.update_value(offset, value)
        return [error]

    @tile_rpc(*rpcs.GET_CONFIG_VARIABLE)
    def get_config_variable(self, config_id, offset):
        """Get a chunk of a config variable's value."""

        config = self._config_variables.get(config_id)
        if config is None:
            return [b""]

        return [bytes(config.current_value[offset:offset + 20])]

    @tile_rpc(*rpcs.TILE_STATUS)
    def tile_status(self):
        """Required status RPC that allows matching a proxy object with a tile."""

        #TODO: Make this status real based on the tile's status

        status = (1 << 1) | (1 << 0)  # Configured and running, not currently used but required for compat with physical tiles
        return [0xFFFF, self.name, 1, 0, 0, status]



TYPE_CODES = {'uint8_t': 'B', 'char': 'B', 'int8_t': 'b', 'uint16_t': 'H', 'int16_t': 'h', 'uint32_t': 'L', 'int32_t': 'l'}


def parse_size_name(type_name):
    """Calculate size and encoding from a type name.

    This method takes a C-style type string like uint8_t[10] and returns
    - the total size in bytes
    - the unit size of each member (if it's an array)
    - the scruct.{pack,unpack} format code for decoding the base type
    - whether it is an array.
    """

    if ' ' in type_name:
        raise ArgumentError("There should not be a space in config variable type specifier", specifier=type_name)

    variable = False
    count = 1
    base_type = type_name

    if type_name[-1] == ']':
        variable = True
        start_index = type_name.find('[')
        if start_index == -1:
            raise ArgumentError("Could not find matching [ for ] character", specifier=type_name)

        count = int(type_name[start_index+1:-1], 0)
        base_type = type_name[:start_index]

    matched_type = TYPE_CODES.get(base_type)
    if matched_type is None:
        raise ArgumentError("Could not find base type name", base_type=base_type, type_string=type_name)

    base_size = struct.calcsize("<%s" % matched_type)
    total_size = base_size*count

    return total_size, base_size, matched_type, variable



"""Subclass of VirtualDeviceAdapter that supports emulated devices.

This DeviceAdapter provides additional DebugManager functionality such as
saving and loading internal state snapshots and loading test scenarios
into the emulated devices.
"""

import logging
from iotile.core.hw.transport import VirtualDeviceAdapter
from iotile.core.hw.exceptions import DeviceAdapterError
from ..virtual import EmulatedDevice


class EmulatedDeviceAdapter(VirtualDeviceAdapter):
    """DeviceAdapter for connecting to EmulatedDevices.

    This adapter is used the exact same way as VirtualDeviceAdapter, except
    that it only supported EmulatedDevice subclasses and it implements the
    debug interface to allow for dumping and loading the state of the
    emulated device.

    Args:
        port (string): A port description that should be in the form of
            device_name1@<optional_config_json1;device_name2@optional_config_json2
        devices (list of EmulatedDevice): Optional list of specific, precreated emulated
            devices that should be added to the device adapter.
    """

    def __init__(self, port, devices=None):
        super(EmulatedDeviceAdapter, self).__init__(port, devices)
        self._logger = logging.getLogger(__name__)

    @classmethod
    def _validate_device(cls, device):
        """Hook for subclases to ensure that only specific kinds of devices are loaded.

        Returns:
            bool: Whether the virtual device is allowed to load.
        """

        return isinstance(device, EmulatedDevice)

    async def send_rpc(self, conn_id, address, rpc_id, payload, timeout):
        """Asynchronously send an RPC to this IOTile device

        Args:
            conn_id (int): A unique identifier that will refer to this connection
            address (int): the address of the tile that we wish to send the RPC to
            rpc_id (int): the 16-bit id of the RPC we want to call
            payload (bytearray): the payload of the command
            timeout (float): the number of seconds to wait for the RPC to execute
        """

        try:
            return await super(EmulatedDeviceAdapter, self).send_rpc(conn_id, address, rpc_id, payload, timeout)
        finally:
            for dev in self.devices.values():
                await dev.wait_idle()

    async def debug(self, conn_id, name, cmd_args):
        """Asynchronously complete a named debug command.

        The command name and arguments are passed to the underlying device adapter
        and interpreted there.

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            name (string): the name of the debug command we want to invoke
            cmd_args (dict): any arguments that we want to send with this command.
        """

        if name == 'heartbeat':
            return {'alive': True}

        device = self._get_property(conn_id, 'device')

        retval = None

        try:
            if name == 'dump_state':
                retval = device.dump_state()
            elif name == 'restore_state':
                state = cmd_args['snapshot']
                device.restore_state(state)
            elif name == 'load_scenario':
                scenario = cmd_args['scenario']
                device.load_metascenario(scenario)
            elif name == 'track_changes':
                if cmd_args['enabled']:
                    device.state_history.enable()
                else:
                    device.state_history.disable()
            elif name == 'dump_changes':
                outpath = cmd_args['path']
                device.state_history.dump(outpath)
            else:
                reason = "Unknown command %s" % name
                raise DeviceAdapterError(conn_id, 'debug {}'.format(name), reason)
        except Exception as exc:
            self._logger.exception("Error processing debug command %s: args=%s", name, cmd_args)
            reason = "Exception %s occurred during processing" % str(exc)
            raise DeviceAdapterError(conn_id, 'debug {}'.format(name), reason) from exc

        return retval



"""RPCs implemented by the ConfigDatabase subsystem on an IOTile controller tile.

These RPCs allow you to persistently store config variables that will be
streamed to specific tiles on each reset.  This is analogous to how
Environment Variables work on unix with the config variable being an
Environment Variable and the tile being a process.

The variables are sent to each tile when the tiles boot and their values do
not change during the lifetime of the tile.
"""

from iotile.core.hw.virtual import RPCDeclaration


START_CONFIG_VAR_ENTRY = RPCDeclaration(0x2a07, "H8s", "L")
"""Start pushing a new config variable to the ConfigDatabase.

The expected use of this RPC is the following sequence:
- START_CONFIG_VAR_ENTRY
- CONTINUE_CONFIG_VAR_ENTRY (1 or more times as needed)
- END_CONFIG_VAR_ENTRY

This series of RPCs will add save config var into the config database.

Args:
  - uint16_t: The config var id that we are sending such as 0x8001
  - char[8]: A serialized SlotIdentifier object that will be used to
    decide which tile(s) this config variable will be sent to.

Returns:
  - uint32_t: An error code.

Possible Errors:
  - STATE_CHANGE_AT_INVALID_TIME: If a config database compaction is in progress.
  - DESTINATION_BUFFER_TOO_SMALL: If there is not enough space for the config variable
    data.
"""

CONTINUE_CONFIG_VAR_ENTRY = RPCDeclaration(0x2a08, "V", "L")
"""Push data to the config variable entry currently in progress.

This RPC should only be called after START_CONFIG_VAR_ENTRY.  The contents
of the RPC call are just a binary blob of data that will be appened to the
current config variable.

You can call this RPC multiple times to push more than 20 bytes of data.

Args:
  - char[]: The binary data that should be appended to the config variable
    in the database.

Returns:
  - uint32_t: An error code.

Possible Errors:
  - DESTINATION_BUFFER_TOO_SMALL: If there is not enough space for the config variable
    data.
"""

END_CONFIG_VAR_ENTRY = RPCDeclaration(0x2a09, "", "L")
"""Finish the current config variable entry.

This RPC is called to commit whatever data has been pushed with previous calls
to CONTINUE_CONFIG_VAR_ENTRY and mark the currently in-progress config var as
complete.

Returns:
  - uint32_t: An error code.

Possible Errors:
  - INPUT_BUFFER_WRONG_SIZE: If the length of the config variable is 0, indicating no previous
    calls to CONFIG_CONFIG_VAR_ENTRY.
"""

GET_CONFIG_VAR_ENTRY = RPCDeclaration(0x2a0a, "H", "L3H8sBB")
"""Get a raw config variable entry by index.

This RPC will return the raw control entry for a config variable
stored in the config database.  The argument is the index of the
variable to get which should be in the range [1, COUNT_CONFIG_VAR_ENTRIES].

Note that the index is one-based since the 0th spot is reserved for magic
number information.

Args:
  - uint16_t: The 1-based index of the config variable entry you want to
    dump.

Returns:
  - uint32_t: An error code.
  - uint16_t: A magic number
  - uint16_t: The raw offset where data starts for this config variable in
    the data storage buffer.
  - uint16_t: The length of data in bytes that is associated with this config
    variable. Note that the config variable id is stored as the first 2 bytes
    of the config data so this will always be greater than 2.
  - char[8]: An encoded SlotIdentifier specifying the target information for
    the config variable.
  - uint8_t: Whether this config entry is still valid.  This will be 0xFF if
    valid.  If not valid, an error will be returned instead so you will never
    see a non-0xFF value here.
  - uint8_t: Reserved, should be ignored.

Possible Errors:
  - STATE_CHANGE_AT_INVALID_TIME: A database compaction is in progress.
  - INVALID_ARRAY_KEY: The index is < 1 or > the number of stored entries.
  - OBSOLETE_ENTRY: The entry has been marked as invalid.
  - INVALID_ENTRY: The entry has an invalid magic number.  This usually indicates
    some kind of database corruption.
"""

COUNT_CONFIG_VAR_ENTRIES = RPCDeclaration(0x2a0b, "", "L")
"""Count the number of stored config variables.

Returns:
  - uint32_t: The count of currently stored config variables including
    variables that have been marked as invalid.  This number will only
    ever decrease after a call to CLEAR_CONFIG_VAR_ENTRIES or
    COMPACT_CONFIG_VAR_ENTRIES.
"""

CLEAR_CONFIG_VAR_ENTRIES = RPCDeclaration(0x2a0c, "", "L")
"""Clear all config variables from the config database.

If database compaction is currently in progress, the function will
return without doing anything and without setting any error code.

Returns:
  - uint32_t: An error code.  No possible errors can be returned
    currently.
"""

GET_CONFIG_VAR_ENTRY_DATA = RPCDeclaration(0x2a0d, "HH", "LV")
"""Get up to 16 bytes of data from a config variable entry.

This RPC take a 1-based index for the variable you want to read
and an offset into the data defined for that variable.  It
returns the next 16-bytes of data starting at that offset.  If
there are less than 16-bytes of data left, it returns less data.

Args:
  - uint16_t: The 1 based index of the config variable we want data from.
  - uint16_t: The offset of the the chunk of data that we want to receive.

Returns:
  - uint32_t: An error code.
  - char[]: Up to 16 bytes of binary data from the selected config variable.

Possible Errors:
  - INVALID_ARRAY_KEY: The index is < 1 or > the number of stored entries or
    the index is valid but the offset is invalid for the length of that entry.
  - OBSOLETE_ENTRY: The entry has been marked as invalid.
  - INVALID_ENTRY: The entry has an invalid magic number.  This usually
    indicates some kind of database corruption.
"""

INVALIDATE_CONFIG_VAR_ENTRY = RPCDeclaration(0x2a0e, "H", "L")
"""Invalidate a config variable entry by index.

This RPC will mark the corresponding config variable as invalid. This
operation effectively erases the config variable but since the data is stored
in an append-only log format, it does not actually free any storage.  The only
way to reclaim the space used by invalidated config entries is by calling the
COMPACT_CONFIG_DATABASE rpc.

Args:
  - uint16_t: The 1 based index of the config variable we want data from.

Returns:
  - uint32_t: An error code.

Possible Errors:
  - INVALID_ARRAY_KEY: The index is < 1 or > the number of stored entries or
    the index is valid but the offset is invalid for the length of that entry.
  - OBSOLETE_ENTRY: The entry has been marked as invalid.
  - INVALID_ENTRY: The entry has an invalid magic number.  This usually
    indicates some kind of database corruption.
"""

COMPACT_CONFIG_DATABASE = RPCDeclaration(0x2a0f, "", "L")
"""Compact the config database by removing all invalid entries.

This RPC will perform garbage collection synchronously on the config database.
Any entries marked as invalid will be removed and the remaining entries will
be shifted left to fill in the space left by the missing entries.

Returns:
  - uint32_t: An error code.  This method cannot fail.
"""

GET_CONFIG_DATABASE_INFO = RPCDeclaration(0x2a10, "", "L6H")
"""Get memory usage and space information about the config database.

This RPC returns a set of useful information about the state of the
config database including how many entries are there, how many are
invalid and how much space count be saved by compacting the database.

Returns:
  - uint32_t: The maximum number of data bytes that can be stored.
  - uint16_t: The current size of all data bytes stored.
  - uint16_t: The current size of all invalid config variables that
    could be saved by compaction.
  - uint16_t: The current count of all stored config variables.
  - uint16_t: The current count of all invalid config variables that
    would be removed by compaction.
  - uint16_t: The maximum number of config variables that can be stored.
  - uint16_t: Reserved.
"""



"""Tests to ensure that our coroutine based emulation loop works."""

import asyncio
import struct
import pytest
from iotile.emulate.internal import EmulationLoop
from iotile.emulate import RPCRuntimeError
from iotile.core.hw.exceptions import AsynchronousRPCResponse
from iotile.core.exceptions import TimeoutExpiredError, InternalError


def test_basic_eventloop():
    """Make sure basic things work."""

    def _rpc_executor(_address, rpc_id, arg_payload):
        if rpc_id == 0x8000:
            raise ValueError("Error")

        return arg_payload

    loop = EmulationLoop(_rpc_executor)
    loop.start()



    try:
        assert loop.call_rpc_external(8, 0x8001, b'abcd') == b'abcd'

        with pytest.raises(ValueError):
            loop.call_rpc_external(8, 0x8000, b'')

    finally:
        loop.stop()


def test_runtime_errors():
    """Make sure runtime errors work."""

    def _rpc_executor(_address, rpc_id, arg_payload):
        if rpc_id == 0x8000:
            raise RPCRuntimeError(0xabcd)

        if rpc_id == 0x8001:
            raise RPCRuntimeError(0xabcd, subsystem=1)

        if rpc_id == 0x8002:
            raise RPCRuntimeError(0xabcd, size="H")

        if rpc_id == 0x8003:
            raise RPCRuntimeError(0xabcd, subsystem=1, size="H")

        return arg_payload

    loop = EmulationLoop(_rpc_executor)
    loop.start()

    try:
        assert loop.call_rpc_external(8, 0x8000, b'abcd') == struct.pack("<L", 0xabcd)
        assert loop.call_rpc_external(8, 0x8001, b'abcd') == struct.pack("<L", 0x1abcd)
        assert loop.call_rpc_external(8, 0x8002, b'abcd') == struct.pack("<H", 0xabcd)

        with pytest.raises(InternalError):
            loop.call_rpc_external(8, 0x8003, b'')

    finally:
        loop.stop()


def test_async_rpc():
    """Make sure we can send an asynchronous RPC."""
    loop = None

    def _rpc_executor(_address, rpc_id, arg_payload):
        if rpc_id == 0x8000:
            raise ValueError("Error")

        if rpc_id == 0x8002:
            address, rpc_id = loop.get_current_rpc()
            asyncio.get_event_loop().call_soon(loop.finish_async_rpc, address, rpc_id, b'4444')
            raise AsynchronousRPCResponse()

        if rpc_id == 0x8003:
            address, rpc_id = loop.get_current_rpc()
            asyncio.get_event_loop().call_soon(loop.finish_async_rpc, address, rpc_id, "L", 0xabcdef00)
            raise AsynchronousRPCResponse()

        return arg_payload

    loop = EmulationLoop(_rpc_executor)
    loop.start()

    try:
        assert loop.call_rpc_external(8, 0x8001, b'abcd') == b'abcd'
        assert loop.call_rpc_external(8, 0x8002, b'abcd') == b'4444'
        assert loop.call_rpc_external(8, 0x8003, b'abcd') == struct.pack("<L", 0xabcdef00)
        loop.wait_idle()

    finally:
        loop.stop()


def test_rpc_timeout():
    """Make sure we can timeout an RPC."""

    def _rpc_executor(_address, rpc_id, arg_payload):
        raise AsynchronousRPCResponse()

    loop = EmulationLoop(_rpc_executor)
    loop.start()

    try:
        with pytest.raises(TimeoutExpiredError):
            loop.call_rpc_external(8, 0x8001, b'abcd', timeout=0.001)

        with pytest.raises(TimeoutExpiredError):
            loop.wait_idle(timeout=0.001)

    finally:
        loop.stop()



"""A simple emulated reference device that includes a blank non-controller tile."""

from ..reference import ReferenceDevice


class DemoReferenceDevice(ReferenceDevice):
    """A basic reference device with a reference controller."""

    pass



"""Base classes for virtual IOTile devices designed to emulate physical IOTile Devices.

Classes that derive from EmulatedDevice and EmulatedTile have additional
funtionality for saving and loading their state to provide for easy creation
of testing scenarios.

For example, you may have an IOTile Device that serves as a shock and
vibration data logger.  In reality it would take time to load it up with real
captured shocks and vibrations in order to test that we could properly create
software that read those waveforms.  With an EmulatedDevice, you can just load
in a `has 100 waveforms` scenario and use that as part of automated
integration test cases.


EmulatedDevice vs VirtualIOTileDevice
=====================================

In general a VirtualIOTileDevice is not designed to emulate a physical device.
A virtual device is a full-fledged IOTile device that just happens to be
running on a normal computer rather than on low-power embedded hardware.

An EmulatedDevice or EmulatedTile is a virtual IOTile device or tile whose
sole purpose in life is to be an emulator for some particular physical tile or
device.

For example, we could have a POD-1M emulator, that would be an EmulatedDevice,
whereas the network configuration device running on an Access Point is just a
normal virtual device.

The idea is that EmulatedTile classes can live with each tile firmware and
EmulatedDevice classes with each production device. The Emulated{Tile,Device}
stays in sync with the features of the physical {tile,device} and provides a
good way to do integration tests of other software components that need to
interact with that physical device.


Lifecycle of an Emulated Device
===============================

Unlike normal virtual devices, which have no specific lifecycle imposed on
them, emulated devices are designed to act as standins for physical IOTile
devices that do need to follow a very specific initialization process.

The process is as follows:

- First the controller tile for the device boots and initializes itself.
- Each peripheral tile checks in with the controller tile and registers itself
    - this triggers the controller tile to stream any config variables that
      pertain to that peripheral tile.
    - once finished, the controller tile sends a start_application RPC to the
      peripheral tile and it begins operation.

In an EmulatedDevice class, this process happens when start() is called on the
device and when the controller is reset().  Each tile can also be reset
independently in order to trigger it to go through its initialization process
again (including registering and receiving new config variables).


Technical Details of Emulation
==============================

All emulation happens inside the EmulatedDevice subclass.  Nothing happens
on __init__ except for internal class initialization.  Emulation begins
when EmulatedDevice.start() is called and finishes when EmulatedDevice.stop()
is called.

Both start() and stop() are synchronous calls in that emulation is guaranteed
to be running when start() returns and guaranteed to be finished when stop()
returns.

Since start() and stop() are synchronous, they are not safe to call from
within the event loop that is used to run the simulation.

Conceptually start() is equivalent to powering on the emulated device and
stop() is equivalent to powering it off.

Emulation is performed inside of a BackgroundEventLoop using
:class:`EmulationLoop`. Each subsystem or tile inside the device that would
normally run in an RTOS task or on a separate processor instead runs as a
coroutine cooperatively scheduled with all other tasks inside the background
event loop.

There is a single background coroutine spawned by EmulatedDevice during the
start() method that is in charge of dispatching rpcs to each tile.  Any time a
tile wishes to communicate with another tile, it calls
``self.device.emulator.await_rpc()`` and this queues the rpc for completion on the
background rpc dispatcher task.  The caller can await that call in order to
yield until the rpc has finished.

There is a clear distinction between routines that are safe to call (by a
user) outside of the emulation loop an those that are not threadsafe and must
not be called outside of the emulation loop.  Some of the convenience methods
declared on ``EmulatedDevice`` can automatically detect what context they are
being called from to know whether they should block or return an awaitable.

This makes it safe to call methods like ``wait_idle``from both inside and outside
of the event loop.  However, some methods are marked to only be safe to call
from outside the event loop like ``EmulatedDevice.rpc()``.

Since all RPCs are guaranteed to only execute on a single thread, only one RPC
at a time can execute, as is the case in a physical IOTile device.

If you want to directly interact with an EmulatedDevice from a control script,
you can call the EmulatedDevice.rpc() method directly from your control
script.

Even though you are synchronously invoking the rpc() method, the rpc that you
specify will still be executed in the background rpc dispatch thread while
your thread blocks waiting for it to finish.

If you want to queue an RPC without blocking, you can use the deferred_rpc()
method, optionally passing a callback that will be invoked when the RPC
finishes.
"""

from .emulated_device import EmulatedDevice
from .peripheral_tile import EmulatedPeripheralTile
from .emulated_tile import EmulatedTile
from .simple_state import SerializableState

__all__ = ['EmulatedDevice', 'EmulatedTile', 'SerializableState', 'EmulatedPeripheralTile']



"""Base class for controller subsystems."""

import asyncio
from iotile.core.exceptions import InternalError

class ControllerSubsystemBase:
    """Base class for Controller subsystems.

    This class allows subsystems to define tasks that run in the background
    and are automatically associated with the parent controller.  It also
    allows the subsytems to update an `initialized` Event that lets the
    controller know when their background tasks are up and running.

    Args:
        emulator (EmulationLoop): The underlying emulation loop that all
            tasks should be added to.
    """

    def __init__(self, emulator):
        self._emulator = emulator
        self.initialized = emulator.create_event()

    async def _reset_vector(self):
        """Initialize any background tasks associated with this subsystem.

        Subclasses that choose to override this method must set the
        self.initialized Event() inside their reset vector once it starts up.
        """

        self.initialized.set()

    def clear_to_reset(self, config_vars):
        """Clear all volatile information across a reset.

        Classes that override this method must call super().clear_to_reset()
        to ensure that their base class is initialized correctly.

        Args:
            config_vars (dict): The map of all config variables that are
                declared on the controller.
        """

        self.initialized.clear()

    async def initialize(self, timeout=2.0):
        """Launch any background tasks associated with this subsystem.

        This method will synchronously await self.initialized() which makes
        sure that the background tasks start up correctly.
        """

        if self.initialized.is_set():
            raise InternalError("initialize called when already initialized")

        self._emulator.add_task(8, self._reset_vector())

        await asyncio.wait_for(self.initialized.wait(), timeout=timeout)



"""Mixin for the raw sensor log.

The raw sensor log is the subsystem that stores sensor readings
on behalf of sensor graph and allows you to query them later.

The main purpose of this subsystem is to serve as a persistent
data store.  Since it allocates `StreamWalker` objects that
are used by the SensorGraph and Streaming subsystems, it needs
to have a very clear dump()/restore() process to ensure that
those subsystems properly get their stream walkers restored
to the correct state.
"""

import struct
import logging
from iotile.core.hw.virtual import tile_rpc
from iotile.core.hw.reports import IOTileReading
from iotile.sg import SensorLog, DataStream, DataStreamSelector
from iotile.sg.engine import InMemoryStorageEngine
from iotile.sg.exceptions import StorageFullError, StreamEmptyError, UnresolvedIdentifierError
from ...constants import rpcs, pack_error, Error, ControllerSubsystem, SensorLogError, streams
from .controller_system import ControllerSubsystemBase

class SensorLogSubsystem(ControllerSubsystemBase):
    """Container for raw sensor log state."""

    def __init__(self, emulator, model):
        super(SensorLogSubsystem, self).__init__(emulator)

        self.engine = InMemoryStorageEngine(model=model)
        self.storage = SensorLog(self.engine, model=model, id_assigner=lambda x, y: self.allocate_id())
        self.dump_walker = None
        self.next_id = 1
        self._logger = logging.getLogger(__name__)

    def dump(self):
        """Serialize the state of this subsystem into a dict.

        Returns:
            dict: The serialized state
        """

        walker = self.dump_walker
        if walker is not None:
            walker = walker.dump()

        state = {
            'storage': self.storage.dump(),
            'dump_walker': walker,
            'next_id': self.next_id
        }

        return state

    def prepare_for_restore(self):
        """Prepare the SensorLog subsystem for a restore.

        This must be called at the start of the restore to clear all of the
        stream walkers in the SensorLog storage engine, which are then added
        back by the SensorGraph and Streaming subsystems and then restored
        to their previous positions when restore() is called on this subsystem.
        """

        self.storage.destroy_all_walkers()

    def restore(self, state):
        """Restore the state of this subsystem from a prior call to dump().

        Calling restore must be properly sequenced with calls to other
        subsystems that include stream walkers so that their walkers are
        properly restored.

        Args:
            state (dict): The results of a prior call to dump().
        """

        self.storage.restore(state.get('storage'))

        dump_walker = state.get('dump_walker')
        if dump_walker is not None:
            dump_walker = self.storage.restore_walker(dump_walker)

        self.dump_walker = dump_walker
        self.next_id = state.get('next_id', 1)

    def clear(self, timestamp):
        """Clear all data from the RSL.

        This pushes a single reading once we clear everything so that
        we keep track of the highest ID that we have allocated to date.

        This needs the current timestamp to be able to properly timestamp
        the cleared storage reading that it pushes.

        Args:
            timestamp (int): The current timestamp to store with the
                reading.
        """

        self.storage.clear()

        self.push(streams.DATA_CLEARED, timestamp, 1)

    def clear_to_reset(self, config_vars):
        """Clear all volatile information across a reset."""

        self._logger.info("Config vars in sensor log reset: %s", config_vars)
        super(SensorLogSubsystem, self).clear_to_reset(config_vars)

        self.storage.destroy_all_walkers()
        self.dump_walker = None

        if config_vars.get('storage_fillstop', False):
            self._logger.debug("Marking storage log fill/stop")
            self.storage.set_rollover('storage', False)

        if config_vars.get('streaming_fillstop', False):
            self._logger.debug("Marking streaming log fill/stop")
            self.storage.set_rollover('streaming', False)

    def count(self):
        """Count many many readings are persistently stored.

        Returns:
            (int, int): The number of readings in storage and output areas.
        """

        return self.storage.count()

    def allocate_id(self):
        """Get the next unique ID.

        Returns:
            int: A unique reading ID.
        """

        next_id = self.next_id
        self.next_id += 1

        return next_id

    def push(self, stream_id, timestamp, value):
        """Push a value to a stream.

        Args:
            stream_id (int): The stream we want to push to.
            timestamp (int): The raw timestamp of the value we want to
                store.
            value (int): The 32-bit integer value we want to push.
        Returns:
            int: Packed 32-bit error code.
        """

        stream = DataStream.FromEncoded(stream_id)
        reading = IOTileReading(stream_id, timestamp, value)

        try:
            self.storage.push(stream, reading)

            return Error.NO_ERROR
        except StorageFullError:
            return pack_error(ControllerSubsystem.SENSOR_LOG, SensorLogError.RING_BUFFER_FULL)

    def inspect_virtual(self, stream_id):
        """Inspect the last value written into a virtual stream.

        Args:
            stream_id (int): The virtual stream was want to inspect.

        Returns:
            (int, int): An error code and the stream value.
        """

        stream = DataStream.FromEncoded(stream_id)

        if stream.buffered:
            return [pack_error(ControllerSubsystem.SENSOR_LOG, SensorLogError.VIRTUAL_STREAM_NOT_FOUND), 0]

        try:
            reading = self.storage.inspect_last(stream, only_allocated=True)
            return [Error.NO_ERROR, reading.value]
        except StreamEmptyError:
            return [Error.NO_ERROR, 0]
        except UnresolvedIdentifierError:
            return [pack_error(ControllerSubsystem.SENSOR_LOG, SensorLogError.VIRTUAL_STREAM_NOT_FOUND), 0]

    def dump_begin(self, selector_id):
        """Start dumping a stream.

        Args:
            selector_id (int): The buffered stream we want to dump.

        Returns:
            (int, int, int): Error code, second error code, number of available readings
        """

        if self.dump_walker is not None:
            self.storage.destroy_walker(self.dump_walker)

        selector = DataStreamSelector.FromEncoded(selector_id)
        self.dump_walker = self.storage.create_walker(selector, skip_all=False)

        return Error.NO_ERROR, Error.NO_ERROR, self.dump_walker.count()

    def dump_seek(self, reading_id):
        """Seek the dump streamer to a given ID.

        Returns:
            (int, int, int): Two error codes and the count of remaining readings.

            The first error code covers the seeking process.
            The second error code covers the stream counting process (cannot fail)
            The third item in the tuple is the number of readings left in the stream.
        """

        if self.dump_walker is None:
            return (pack_error(ControllerSubsystem.SENSOR_LOG, SensorLogError.STREAM_WALKER_NOT_INITIALIZED),
                    Error.NO_ERROR, 0)

        try:
            exact = self.dump_walker.seek(reading_id, target='id')
        except UnresolvedIdentifierError:
            return (pack_error(ControllerSubsystem.SENSOR_LOG, SensorLogError.NO_MORE_READINGS),
                    Error.NO_ERROR, 0)

        error = Error.NO_ERROR
        if not exact:
            error = pack_error(ControllerSubsystem.SENSOR_LOG, SensorLogError.ID_FOUND_FOR_ANOTHER_STREAM)

        return (error, Error.NO_ERROR, self.dump_walker.count())

    def dump_next(self):
        """Dump the next reading from the stream.

        Returns:
            IOTileReading: The next reading or None if there isn't one
        """

        if self.dump_walker is None:
            return pack_error(ControllerSubsystem.SENSOR_LOG, SensorLogError.STREAM_WALKER_NOT_INITIALIZED)

        try:
            return self.dump_walker.pop()
        except StreamEmptyError:
            return None

    def highest_stored_id(self):
        """Scan through the stored readings and report the highest stored id.

        Returns:
            int: The highest stored id.
        """

        shared = [0]
        def _keep_max(_i, reading):
            if reading.reading_id > shared[0]:
                shared[0] = reading.reading_id

        self.engine.scan_storage('storage', _keep_max)
        self.engine.scan_storage('streaming', _keep_max)

        return shared[0]


class RawSensorLogMixin(object):
    """Reference controller subsystem for the raw sensor log.

    This class must be used as a mixin with a ReferenceController base class.

    Args:
        model (DeviceModel): The device model to use to calculate
            constraints and other operating parameters.
    """


    def __init__(self, emulator, model):
        self.sensor_log = SensorLogSubsystem(emulator, model)
        self._post_config_subsystems.append(self.sensor_log)

        # Declare all of our config variables
        self.declare_config_variable('storage_fillstop', 0x2004, 'uint8_t', default=False, convert='bool')
        self.declare_config_variable('streaming_fillstop', 0x2005, 'uint8_t', default=False, convert='bool')


    @tile_rpc(*rpcs.RSL_PUSH_READING)
    def rsl_push_reading(self, value, stream_id):
        """Push a reading to the RSL directly."""

        #FIXME: Fix this with timestamp from clock manager task
        err = self.sensor_log.push(stream_id, 0, value)
        return [err]

    @tile_rpc(*rpcs.RSL_PUSH_MANY_READINGS)
    def rsl_push_many_readings(self, value, count, stream_id):
        """Push many copies of a reading to the RSL."""

        #FIXME: Fix this with timestamp from clock manager task

        for i in range(1, count+1):
            err = self.sensor_log.push(stream_id, 0, value)
            if err != Error.NO_ERROR:
                return [err, i]

        return [Error.NO_ERROR, count]

    @tile_rpc(*rpcs.RSL_COUNT_READINGS)
    def rsl_count_readings(self):
        """Count how many readings are stored in the RSL."""

        storage, output = self.sensor_log.count()
        return [Error.NO_ERROR, storage, output]

    @tile_rpc(*rpcs.RSL_CLEAR_READINGS)
    def rsl_clear_readings(self):
        """Clear all data from the RSL."""

        #FIXME: Fix this with timestamp from clock manager task
        self.sensor_log.clear(0)
        return [Error.NO_ERROR]

    @tile_rpc(*rpcs.RSL_INSPECT_VIRTUAL_STREAM)
    def rsl_inspect_virtual_stream(self, stream_id):
        """Inspect the last value in a virtual stream."""

        return self.sensor_log.inspect_virtual(stream_id)

    @tile_rpc(*rpcs.RSL_DUMP_STREAM_BEGIN)
    def rsl_dump_stream_begin(self, stream_id):
        """Begin dumping the contents of a stream."""

        err, err2, count = self.sensor_log.dump_begin(stream_id)

        #FIXME: Fix this with the uptime of the clock manager task
        return [err, err2, count, 0]

    @tile_rpc(*rpcs.RSL_DUMP_STREAM_SEEK)
    def rsl_dump_stream_seek(self, reading_id):
        """Seek a specific reading by ID."""

        return self.sensor_log.dump_seek(reading_id)

    @tile_rpc(*rpcs.RSL_DUMP_STREAM_NEXT)
    def rsl_dump_stream_next(self, output_format):
        """Dump the next reading from the output stream."""

        timestamp = 0
        stream_id = 0
        value = 0
        reading_id = 0
        error = Error.NO_ERROR

        reading = self.sensor_log.dump_next()
        if reading is not None:
            timestamp = reading.raw_time
            stream_id = reading.stream
            value = reading.value
            reading_id = reading.reading_id
        else:
            error = pack_error(ControllerSubsystem.SENSOR_LOG, SensorLogError.NO_MORE_READINGS)

        if output_format == 0:
            return [struct.pack("<LLL", error, timestamp, value)]
        elif output_format != 1:
            raise ValueError("Output format other than 1 not yet supported")

        return [struct.pack("<LLLLH2x", error, timestamp, value, reading_id, stream_id)]

    @tile_rpc(*rpcs.RSL_HIGHEST_READING_ID)
    def rsl_get_highest_saved_id(self):
        """Get the highest saved reading id."""

        highest_id = self.sensor_log.highest_stored_id()
        return [Error.NO_ERROR, highest_id]



"""Mixin classes that implement specific controller subsystems.

Since the IOTile controller has a lot of different subsystems, it's more
clear to break each one up as a separate mixin class to its easier to
see how each bit of functionality relates.

The specific API and semantics of a controller mixin is as follows:
- The mixin must start in a reset state when it is created.
- The mixin must support a dump() function that serializes its current state
  to a dict.
- The mixin must support a restore(state) function that takes a dict state
  produced by a prior call to dump() and loads it.
- If the mixin needs to reinitialize itself when a controller reset happens
  it needs to implement a clear_to_reset(config_vars) method that gets a
  copy of all of the config variables defined for the tile.  It also needs
  to add itself to the _post_config_subsystems array in the mixin __init__
  function.
"""

from .remote_bridge import RemoteBridgeMixin
from .tile_manager import TileManagerMixin
from .config_database import ConfigDatabaseMixin
from .sensor_log import RawSensorLogMixin
from .sensor_graph import SensorGraphMixin
from .stream_manager import StreamingSubsystemMixin
from .clock_manager import ClockManagerMixin

__all__ = ['RemoteBridgeMixin', 'TileManagerMixin', 'ConfigDatabaseMixin',
           'RawSensorLogMixin', 'SensorGraphMixin', 'StreamingSubsystemMixin',
           'ClockManagerMixin']



"""Globally defined stream names.

These stream declarations document the major system streams that are used
internally in an IOTile device to log important system events.  Each
declaration includes a description of how to interpret the contents of data
logged to that stream.

Most of these streams are system inputs, which means they are presented to the
running sensor-graph for actions but ignored by default if no actions react to
them.
"""

from iotile.sg import DataStream

def declare_stream(string_name):
    """Create a StreamDeclaration from a string name.

    This will encode the string name into a 16-bit stream
    identifier.

    Args:
        string_name (str): The human-readable name of the stream.

    Returns:
        int: The stream declaration.
"""

    return DataStream.FromString(string_name).encode()


SYSTEM_RESET = declare_stream('system input 1024')
"""Logs a reading every time the IOTile device is reset with SensorGraph running.

This stream records system reboots.  It is only used when there is a
programmed sensor_graph script running in the device.  It can be used for
debugging if a system reboot is unexpected or for keeping track of when the
device's internal uptime is expected to reset back to zero.

The contents of each reading logged in this stream are an oquare integer value
that is architecture specific and describes the technical cause of the reset
(such as power on, brownout, etc).
"""

NORMAL_TICK = declare_stream('system input 2')
"""Sent every 10 seconds to trigger periodic sensor graph rules.

This is the base tick that forms that periodic input for most long running
timers inside the sensor_graph subsystem.  It cannot be turned off or
configured.
"""

FAST_TICK = declare_stream('system input 3')
"""A fast 1 second tick that is disabled by default.

Technically this tick can be user controlled and set to a value other than one
second but it is controlled internally by the sensor-graph compiler so if you
want to use it for another purpose, you need to make sure it is not referenced
by any rules generated by your .sgf file.
"""

USER_TICK_1 = declare_stream('system input 5')
"""A user controlled tick.

This tick can take any integer second value.  It defaults to being disabled.
These ticks are most valuable for implementing field-programmable timers that
are easy to change with a simple RPC.
"""

USER_TICK_2 = declare_stream('system input 6')
"""A user controlled tick.

This tick can take any integer second value.  It defaults to being disabled.
These ticks are most valuable for implementing field-programmable timers that
are easy to change with a simple RPC.
"""

COMM_TILE_OPEN = declare_stream('system input 1025')
"""Sent when a communications tile opens its streaming interface.

The value in the stream is the address of the tile that opened its streaming
interface.  This input to sensor_graph can be used for entering high-power or
high-speed modes to send more data when there is a user actively connected to
the device.
"""

COMM_TILE_CLOSED = declare_stream('system input 1026')
"""Sent when a communications tile closes its streaming interface.

The value in the stream is the address of the tile the closed its streaming
interface.  This input can be used for leaving high-power or high-speed data
collection modes since there is not longer a user actively connected to the
device.
"""

DATA_CLEARED = declare_stream('system output 1027')
"""Logs a reading every time the sensor log subsystem receives a clear() command.

Each entry in this data stream marks the point in time where the sensor log
was cleared.  It can be used for debugging or forensic purposes if such a
clear was unexpected.  Internally it is used to ensure that there is always a
valid reading inside the sensor log to store the current highest allocated
reading id.
"""



"""Well-known global error codes that can be returned by emulated RPCs."""

from enum import IntEnum

class Error(IntEnum):
    """All globally defined common errors that an RPC may report."""

    NO_ERROR = 0
    """No error occurred."""

    STATIC_MEMORY_ALLOCATION_ERROR = 1
    """There was an error allocating static memory."""

    ERROR_IN_ROM_DRIVER = 2
    """There was an error inside a third party ROM driver."""

    INPUT_BUFFER_TOO_LONG = 3
    """The input buffer was too long."""

    SETUP_FUNCTION_PERFORMED_TWICE = 4
    """A setup function was performed twice."""

    REQUIRED_SETUP_NOT_PERFORMED = 5
    """A required setup function was not performed."""

    INVALID_ARRAY_KEY = 6
    """An invalid array key was passed."""

    INVALID_CHECKSUM = 7
    """An object had an invalid checksum."""

    INVALID_OFFSET = 8
    """An invalid offset was given."""

    STATE_CHANGE_AT_INVALID_TIME = 9
    """The state of an object changed at an unexpected time."""

    TIMEOUT_ERROR = 10
    """An operation has timed out."""

    UNIMPLEMENTED_ERROR = 11
    """The requested operation is not (yet) implemented."""

    UNKNOWN_ERROR = 12
    """An unknown error occurred."""

    HARDWARE_PERIPHERAL_ERROR = 13
    """An internal hardware peripheral inside the microcontroller or SOC had a failure."""

    EXTERNAL_HARDWARE_ERROR = 14
    """An external hardware device had an unrecoverable failure."""

    INVALID_UNALIGNED_ADDRESS = 15
    """A request was made for for an unaligned memory address."""

    DESTINATION_BUFFER_TOO_SMALL = 16
    """The destination buffer is too small."""

    INPUT_BUFFER_WRONG_SIZE = 17
    """The input buffer passed was the wrong size."""

    LOW_VOLTAGE_WARNING = 18
    """An observed voltage triggered a low-voltage warning."""

    HIGH_VOLTAGE_WARNING = 19
    """An observed voltage triggered a high-voltage warning."""

    INVALID_VOLTAGE_ERROR = 20
    """An observed voltage was invalid."""

    INPUT_BUFFER_TOO_SMALL = 21
    """The input buffer passed was too small."""

    INVALID_MAGIC_NUMBER = 22
    """An magic number did not match its expected value."""

    INVALID_HASH_VALUE = 23
    """A calculated hash value did not match its expected value."""

    UNALIGNED_VARIABLE_ERROR = 24
    """A variable was not properly aligned."""


class ConfigDatabaseError(IntEnum):
    """These are subsystem specific error codes potentially returned by the ConfigDatabase."""

    OBSOLETE_ENTRY = 0x8000
    """The config variable entry requested has been invalidated."""

    VARIABLE_DOES_NOT_MATCH = 0x8001
    """The requested config variable does not match."""

    INVALID_ENTRY = 0x8002
    """The requested config variable entry has an invalid magic number."""


class SensorLogError(IntEnum):
    """These are subsystem specific error codes potentially returned by the SensorLog."""

    NO_MORE_READINGS = 0x8000
    """There are no more available readings."""

    VIRTUAL_STREAM_NOT_FOUND = 0x8001
    """There is no stream walker allocated for a specific virtual stream."""

    STREAM_WALKER_NOT_FOUND = 0x8002
    """Could not find the desired stream walker."""

    IMMEDIATE_VALUE_NOT_SUPPORTED = 0x8003
    """This stream does not support immediate values."""

    CANNOT_SKIP_INEXHAUSTIBLE_STREAM = 0x8004  #pylint:disable=invalid-name;This is the name.
    """You cannot call skip on a constant stream that is infinitely deep."""

    STREAM_WALKER_NOT_INITIALIZED = 0x8005
    """The stream walker structure was not properly initialized."""

    CANNOT_USE_UNBUFFERED_STREAM = 0x8006
    """This method cannot use an unbuffered stream."""

    ID_FOUND_FOR_ANOTHER_STREAM = 0x8007
    """Seek_id was called and the id was found but belonged to another stream."""

    RING_BUFFER_FULL = 0x8008
    """The RSL is in fill-stop mode and the desired storage area is full."""


class SensorGraphError(IntEnum):
    """These are subsystem specific error coded potentially returned by SensorGraph."""

    INVALID_NODE_INDEX = 0x8000

    UNINITIALIZED_NODE = 0x8001

    NODE_ALREADY_INITIALIZED = 0x8002

    INVALID_INPUT_INDEX = 0x8003

    NO_PROCESSING_FUNCTION = 0x8004

    NO_UPDATED_OUTPUT = 0x8005

    GRAPH_OFFLINE = 0x8006

    TASK_ALREADY_PENDING = 0x8007

    READING_IGNORED = 0x8008

    INVALID_PROCESSING_FUNCTION = 0x8009

    NO_AVAILABLE_OUTPUTS = 0x800A

    STREAM_NOT_IN_USE = 0x800B

    NO_NODE_SPACE_AVAILABLE = 0x800C

    INVALID_NODE_STREAM = 0x800D

    STREAM_ALREADY_IN_USE = 0x800E

    NODE_NOT_TRIGGERED = 0x800F

    INVALID_STREAM_DESTINATION = 0x8010

    STREAM_ALREADY_IN_PROGRESS = 0x8011

    ERROR_STARTING_STREAMING = 0x8012

    NO_MORE_STREAMER_RESOURCES = 0x8013

    NO_PERSISTED_GRAPH = 0x8014

    OLD_PERSISTED_GRAPH = 0x8015

    INVALID_PERSISTED_GRAPH = 0x8016

    INVALID_BACKOFF_STRATEGY = 0x8017

    STREAMER_NOT_ALLOCATED = 0x8018

    BACKOFF_IN_PROGRESS = 0x8019

    NO_MORE_SELECTOR_NAMES = 0x801A

    UNKNOWN_SELECTOR_NAME = 0x801B

    SELECTOR_DOES_NOT_MATCH_MODULE = 0x801C

    TOO_MANY_ROOT_NODES = 0x801D

    OLD_ACKNOWLEDGE_UPDATE = 0x801E

    STREAMER_HAS_NO_NEW_DATA = 0x801F

    STREAMER_NOT_TRIGGERED = 0x8020
